# coding: utf8
# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict

import paddle.fluid as fluid
from .model_utils.libs import scope, name_scope
from .model_utils.libs import bn, bn_relu, relu, qsigmoid
from .model_utils.libs import conv, max_pool, deconv
from .model_utils.libs import separate_conv
from .model_utils.libs import sigmoid_to_softmax
from .model_utils.loss import softmax_with_loss
from .model_utils.loss import dice_loss
from .model_utils.loss import bce_loss
from paddlex.cv.nets.xception import Xception
from paddlex.cv.nets.mobilenet_v2 import MobileNetV2


class DeepLabv3p(object):
    """实现DeepLabv3+模型
    `"Encoder-Decoder with Atrous Separable Convolution for Semantic Image Segmentation"
    <https://arxiv.org/abs/1802.02611>`

    Args:
        num_classes (int): 类别数。
        backbone (paddlex.cv.nets): 神经网络，实现DeepLabv3+特征图的计算。
        mode (str): 网络运行模式，根据mode构建网络的输入和返回。
            当mode为'train'时，输入为image(-1, 3, -1, -1)和label (-1, 1, -1, -1) 返回loss。
            当mode为'train'时，输入为image (-1, 3, -1, -1)和label  (-1, 1, -1, -1)，返回loss，
            pred (与网络输入label 相同大小的预测结果，值代表相应的类别），label，mask（非忽略值的mask，
            与label相同大小，bool类型）。
            当mode为'test'时，输入为image(-1, 3, -1, -1)返回pred (-1, 1, -1, -1)和
            logit (-1, num_classes, -1, -1) 通道维上代表每一类的概率值。
        output_stride (int): backbone 输出特征图相对于输入的下采样倍数，一般取值为8或16。
        aspp_with_sep_conv (bool): 在asspp模块是否采用separable convolutions。
        decoder_use_sep_conv (bool)： decoder模块是否采用separable convolutions。
        encoder_with_aspp (bool): 是否在encoder阶段采用aspp模块。
        enable_decoder (bool): 是否使用decoder模块。
        use_bce_loss (bool): 是否使用bce loss作为网络的损失函数，只能用于两类分割。可与dice loss同时使用。
        use_dice_loss (bool): 是否使用dice loss作为网络的损失函数，只能用于两类分割，可与bce loss同时使用。
            当use_bce_loss和use_dice_loss都为False时，使用交叉熵损失函数。
        class_weight (list/str): 交叉熵损失函数各类损失的权重。当class_weight为list的时候，长度应为
            num_classes。当class_weight为str时， weight.lower()应为'dynamic'，这时会根据每一轮各类像素的比重
            自行计算相应的权重，每一类的权重为：每类的比例 * num_classes。class_weight取默认值None是，各类的权重1，
            即平时使用的交叉熵损失函数。
        ignore_index (int): label上忽略的值，label为ignore_index的像素不参与损失函数的计算。
        fixed_input_shape (list): 长度为2，维度为1的list，如:[640,720]，用来固定模型输入:'image'的shape，默认为None。

    Raises:
        ValueError: use_bce_loss或use_dice_loss为真且num_calsses > 2。
        ValueError: class_weight为list, 但长度不等于num_class。
            class_weight为str, 但class_weight.low()不等于dynamic。
        TypeError: class_weight不为None时，其类型不是list或str。
    """

    def __init__(self,
                 num_classes,
                 backbone,
                 input_channel=3,
                 mode='train',
                 output_stride=16,
                 aspp_with_sep_conv=True,
                 decoder_use_sep_conv=True,
                 encoder_with_aspp=True,
                 enable_decoder=True,
                 use_bce_loss=False,
                 use_dice_loss=False,
                 class_weight=None,
                 ignore_index=255,
                 fixed_input_shape=None,
                 pooling_stride=[1, 1],
                 pooling_crop_size=None,
                 aspp_with_se=False,
                 se_use_qsigmoid=False,
                 aspp_convs_filters=256,
                 aspp_with_concat_projection=True,
                 add_image_level_feature=True,
                 use_sum_merge=False,
                 conv_filters=256,
                 output_is_logits=False):
        # dice_loss或bce_loss只适用两类分割中
        if num_classes > 2 and (use_bce_loss or use_dice_loss):
            raise ValueError(
                "dice loss and bce loss is only applicable to binary classfication"
            )

        if class_weight is not None:
            if isinstance(class_weight, list):
                if len(class_weight) != num_classes:
                    raise ValueError(
                        "Length of class_weight should be equal to number of classes"
                    )
            elif isinstance(class_weight, str):
                if class_weight.lower() != 'dynamic':
                    raise ValueError(
                        "if class_weight is string, must be dynamic!")
            else:
                raise TypeError(
                    'Expect class_weight is a list or string but receive {}'.
                    format(type(class_weight)))

        self.num_classes = num_classes
        self.input_channel = input_channel
        self.backbone = backbone
        self.mode = mode
        self.use_bce_loss = use_bce_loss
        self.use_dice_loss = use_dice_loss
        self.class_weight = class_weight
        self.ignore_index = ignore_index
        self.output_stride = output_stride
        self.aspp_with_sep_conv = aspp_with_sep_conv
        self.decoder_use_sep_conv = decoder_use_sep_conv
        self.encoder_with_aspp = encoder_with_aspp
        self.enable_decoder = enable_decoder
        self.fixed_input_shape = fixed_input_shape
        self.output_is_logits = output_is_logits
        self.aspp_convs_filters = aspp_convs_filters
        self.output_stride = output_stride
        self.pooling_crop_size = pooling_crop_size
        self.pooling_stride = pooling_stride
        self.se_use_qsigmoid = se_use_qsigmoid
        self.aspp_with_concat_projection = aspp_with_concat_projection
        self.add_image_level_feature = add_image_level_feature
        self.aspp_with_se = aspp_with_se
        self.use_sum_merge = use_sum_merge
        self.conv_filters = conv_filters

    def _encoder(self, input):
        # 编码器配置，采用ASPP架构，pooling + 1x1_conv + 三个不同尺度的空洞卷积并行, concat后1x1conv
        # ASPP_WITH_SEP_CONV：默认为真，使用depthwise可分离卷积，否则使用普通卷积
        # OUTPUT_STRIDE: 下采样倍数，8或16，决定aspp_ratios大小
        # aspp_ratios：ASPP模块空洞卷积的采样率

        if self.output_stride == 16:
            aspp_ratios = [6, 12, 18]
        elif self.output_stride == 8:
            aspp_ratios = [12, 24, 36]
        else:
            aspp_ratios = []

        param_attr = fluid.ParamAttr(
            name=name_scope + 'weights',
            regularizer=None,
            initializer=fluid.initializer.TruncatedNormal(
                loc=0.0, scale=0.06))

        concat_logits = []
        with scope('encoder'):
            channel = self.aspp_convs_filters
            with scope("image_pool"):
                if self.pooling_crop_size is None:
                    image_avg = fluid.layers.reduce_mean(
                        input, [2, 3], keep_dim=True)
                else:
                    pool_w = int((self.pooling_crop_size[0] - 1.0) /
                                 self.output_stride + 1.0)
                    pool_h = int((self.pooling_crop_size[1] - 1.0) /
                                 self.output_stride + 1.0)
                    image_avg = fluid.layers.pool2d(
                        input,
                        pool_size=(pool_h, pool_w),
                        pool_stride=self.pooling_stride,
                        pool_type='avg',
                        pool_padding='VALID')

                act = qsigmoid if self.se_use_qsigmoid else bn_relu

                image_avg = act(
                    conv(
                        image_avg,
                        channel,
                        1,
                        1,
                        groups=1,
                        padding=0,
                        param_attr=param_attr))
                input_shape = fluid.layers.shape(input)
                image_avg = fluid.layers.resize_bilinear(image_avg,
                                                         input_shape[2:])
                if self.add_image_level_feature:
                    concat_logits.append(image_avg)

            with scope("aspp0"):
                aspp0 = bn_relu(
                    conv(
                        input,
                        channel,
                        1,
                        1,
                        groups=1,
                        padding=0,
                        param_attr=param_attr))
                concat_logits.append(aspp0)

            if aspp_ratios:
                with scope("aspp1"):
                    if self.aspp_with_sep_conv:
                        aspp1 = separate_conv(
                            input,
                            channel,
                            1,
                            3,
                            dilation=aspp_ratios[0],
                            act=relu)
                    else:
                        aspp1 = bn_relu(
                            conv(
                                input,
                                channel,
                                stride=1,
                                filter_size=3,
                                dilation=aspp_ratios[0],
                                padding=aspp_ratios[0],
                                param_attr=param_attr))
                    concat_logits.append(aspp1)
                with scope("aspp2"):
                    if self.aspp_with_sep_conv:
                        aspp2 = separate_conv(
                            input,
                            channel,
                            1,
                            3,
                            dilation=aspp_ratios[1],
                            act=relu)
                    else:
                        aspp2 = bn_relu(
                            conv(
                                input,
                                channel,
                                stride=1,
                                filter_size=3,
                                dilation=aspp_ratios[1],
                                padding=aspp_ratios[1],
                                param_attr=param_attr))
                    concat_logits.append(aspp2)
                with scope("aspp3"):
                    if self.aspp_with_sep_conv:
                        aspp3 = separate_conv(
                            input,
                            channel,
                            1,
                            3,
                            dilation=aspp_ratios[2],
                            act=relu)
                    else:
                        aspp3 = bn_relu(
                            conv(
                                input,
                                channel,
                                stride=1,
                                filter_size=3,
                                dilation=aspp_ratios[2],
                                padding=aspp_ratios[2],
                                param_attr=param_attr))
                    concat_logits.append(aspp3)

            with scope("concat"):
                data = fluid.layers.concat(concat_logits, axis=1)
                if self.aspp_with_concat_projection:
                    data = bn_relu(
                        conv(
                            data,
                            channel,
                            1,
                            1,
                            groups=1,
                            padding=0,
                            param_attr=param_attr))
                    data = fluid.layers.dropout(data, 0.9)
            if self.aspp_with_se:
                data = data * image_avg
            return data

    def _decoder_with_sum_merge(self, encode_data, decode_shortcut,
                                param_attr):
        decode_shortcut_shape = fluid.layers.shape(decode_shortcut)
        encode_data = fluid.layers.resize_bilinear(encode_data,
                                                   decode_shortcut_shape[2:])

        encode_data = conv(
            encode_data,
            self.conv_filters,
            1,
            1,
            groups=1,
            padding=0,
            param_attr=param_attr)

        with scope('merge'):
            decode_shortcut = conv(
                decode_shortcut,
                self.conv_filters,
                1,
                1,
                groups=1,
                padding=0,
                param_attr=param_attr)

            return encode_data + decode_shortcut

    def _decoder_with_concat(self, encode_data, decode_shortcut, param_attr):
        with scope('concat'):
            decode_shortcut = bn_relu(
                conv(
                    decode_shortcut,
                    48,
                    1,
                    1,
                    groups=1,
                    padding=0,
                    param_attr=param_attr))

            decode_shortcut_shape = fluid.layers.shape(decode_shortcut)
            encode_data = fluid.layers.resize_bilinear(
                encode_data, decode_shortcut_shape[2:])
            encode_data = fluid.layers.concat(
                [encode_data, decode_shortcut], axis=1)
        if self.decoder_use_sep_conv:
            with scope("separable_conv1"):
                encode_data = separate_conv(
                    encode_data, self.conv_filters, 1, 3, dilation=1, act=relu)
            with scope("separable_conv2"):
                encode_data = separate_conv(
                    encode_data, self.conv_filters, 1, 3, dilation=1, act=relu)
        else:
            with scope("decoder_conv1"):
                encode_data = bn_relu(
                    conv(
                        encode_data,
                        self.conv_filters,
                        stride=1,
                        filter_size=3,
                        dilation=1,
                        padding=1,
                        param_attr=param_attr))
            with scope("decoder_conv2"):
                encode_data = bn_relu(
                    conv(
                        encode_data,
                        self.conv_filters,
                        stride=1,
                        filter_size=3,
                        dilation=1,
                        padding=1,
                        param_attr=param_attr))
        return encode_data

    def _decoder(self, encode_data, decode_shortcut):
        # 解码器配置
        # encode_data：编码器输出
        # decode_shortcut: 从backbone引出的分支, resize后与encode_data concat
        # decoder_use_sep_conv: 默认为真，则concat后连接两个可分离卷积，否则为普通卷积
        param_attr = fluid.ParamAttr(
            name=name_scope + 'weights',
            regularizer=None,
            initializer=fluid.initializer.TruncatedNormal(
                loc=0.0, scale=0.06))

        with scope('decoder'):
            if self.use_sum_merge:
                return self._decoder_with_sum_merge(
                    encode_data, decode_shortcut, param_attr)

            return self._decoder_with_concat(encode_data, decode_shortcut,
                                             param_attr)

    def _get_loss(self, logit, label, mask):
        avg_loss = 0
        if not (self.use_dice_loss or self.use_bce_loss):
            avg_loss += softmax_with_loss(
                logit,
                label,
                mask,
                num_classes=self.num_classes,
                weight=self.class_weight,
                ignore_index=self.ignore_index)
        else:
            if self.use_dice_loss:
                avg_loss += dice_loss(logit, label, mask)
            if self.use_bce_loss:
                avg_loss += bce_loss(
                    logit, label, mask, ignore_index=self.ignore_index)

        return avg_loss

    def generate_inputs(self):
        inputs = OrderedDict()

        if self.fixed_input_shape is not None:
            input_shape = [
                None, self.input_channel, self.fixed_input_shape[1],
                self.fixed_input_shape[0]
            ]
            inputs['image'] = fluid.data(
                dtype='float32', shape=input_shape, name='image')
        else:
            inputs['image'] = fluid.data(
                dtype='float32',
                shape=[None, self.input_channel, None, None],
                name='image')
        if self.mode == 'train':
            inputs['label'] = fluid.data(
                dtype='int32', shape=[None, 1, None, None], name='label')
        return inputs

    def build_net(self, inputs):
        # 在两类分割情况下，当loss函数选择dice_loss或bce_loss的时候，最后logit输出通道数设置为1
        if self.use_dice_loss or self.use_bce_loss:
            self.num_classes = 1
        image = inputs['image']

        if 'MobileNetV3' in self.backbone.__class__.__name__:
            data, decode_shortcut = self.backbone(image)
        else:
            data, decode_shortcuts = self.backbone(image)
            decode_shortcut = decode_shortcuts[self.backbone.decode_points]

        # 编码器解码器设置
        if self.encoder_with_aspp:
            data = self._encoder(data)
        if self.enable_decoder:
            data = self._decoder(data, decode_shortcut)

        # 根据类别数设置最后一个卷积层输出，并resize到图片原始尺寸
        param_attr = fluid.ParamAttr(
            name=name_scope + 'weights',
            regularizer=fluid.regularizer.L2DecayRegularizer(
                regularization_coeff=0.0),
            initializer=fluid.initializer.TruncatedNormal(
                loc=0.0, scale=0.01))
        if not self.output_is_logits:
            with scope('logit'):
                with fluid.name_scope('last_conv'):
                    logit = conv(
                        data,
                        self.num_classes,
                        1,
                        stride=1,
                        padding=0,
                        bias_attr=True,
                        param_attr=param_attr)
        else:
            logit = data

        image_shape = fluid.layers.shape(image)
        logit = fluid.layers.resize_bilinear(logit, image_shape[2:])

        if self.num_classes == 1:
            out = sigmoid_to_softmax(logit)
            out = fluid.layers.transpose(out, [0, 2, 3, 1])
        else:
            out = fluid.layers.transpose(logit, [0, 2, 3, 1])

        pred = fluid.layers.argmax(out, axis=3)
        pred = fluid.layers.unsqueeze(pred, axes=[3])

        if self.mode == 'train':
            label = inputs['label']
            mask = label != self.ignore_index
            return self._get_loss(logit, label, mask)

        elif self.mode == 'eval':
            label = inputs['label']
            mask = label != self.ignore_index
            loss = self._get_loss(logit, label, mask)
            return loss, pred, label, mask
        else:
            if self.num_classes == 1:
                logit = sigmoid_to_softmax(logit)
            else:
                logit = fluid.layers.softmax(logit, axis=1)
            return pred, logit

        return logit



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path as osp
import cv2
from PIL import Image
import numpy as np
import math
from .imgaug_support import execute_imgaug
from .cls_transforms import ClsTransform
from .det_transforms import DetTransform
from .seg_transforms import SegTransform
import paddlex as pdx
from paddlex.cv.models.utils.visualize import get_color_map_list


def _draw_rectangle_and_cname(img, xmin, ymin, xmax, ymax, cname, color):
    """ 根据提供的标注信息，给图片描绘框体和类别显示

    Args:
        img: 图片路径
        xmin: 检测框最小的x坐标
        ymin: 检测框最小的y坐标
        xmax: 检测框最大的x坐标
        ymax: 检测框最大的y坐标
        cname: 类别信息
        color: 类别与颜色的对应信息
    """
    # 描绘检测框
    line_width = math.ceil(2 * max(img.shape[0:2]) / 600)
    cv2.rectangle(
        img,
        pt1=(xmin, ymin),
        pt2=(xmax, ymax),
        color=color,
        thickness=line_width)
    return img


def cls_compose(im, label=None, transforms=None, vdl_writer=None, step=0):
    """
        Args:
            im (str/np.ndarray): 图像路径/图像np.ndarray数据。
            label (int): 每张图像所对应的类别序号。
            vdl_writer (visualdl.LogWriter): VisualDL存储器，日志信息将保存在其中。
                当为None时，不对日志进行保存。默认为None。
            step (int): 数据预处理的轮数，当vdl_writer不为None时有效。默认为0。

        Returns:
            tuple: 根据网络所需字段所组成的tuple；
                字段由transforms中的最后一个数据预处理操作决定。
        """
    if isinstance(im, np.ndarray):
        if len(im.shape) != 3:
            raise Exception(
                "im should be 3-dimension, but now is {}-dimensions".format(
                    len(im.shape)))
    else:
        try:
            im = cv2.imread(im).astype('float32')
        except:
            raise TypeError('Can\'t read The image file {}!'.format(im))
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    if vdl_writer is not None:
        vdl_writer.add_image(
            tag='0. OriginalImage/' + str(step), img=im, step=0)
    op_id = 1
    for op in transforms:
        if isinstance(op, ClsTransform):
            if vdl_writer is not None and hasattr(op, 'prob'):
                op.prob = 1.0
            outputs = op(im, label)
            im = outputs[0]
            if len(outputs) == 2:
                label = outputs[1]
            if isinstance(op, pdx.cv.transforms.cls_transforms.Normalize):
                continue
        else:
            import imgaug.augmenters as iaa
            if isinstance(op, iaa.Augmenter):
                im = execute_imgaug(op, im)
            outputs = (im, )
            if label is not None:
                outputs = (im, label)
        if vdl_writer is not None:
            tag = str(op_id) + '. ' + op.__class__.__name__ + '/' + str(step)
            vdl_writer.add_image(tag=tag, img=im, step=0)
        op_id += 1


def det_compose(im,
                im_info=None,
                label_info=None,
                transforms=None,
                vdl_writer=None,
                step=0,
                labels=[],
                catid2color=None):
    def decode_image(im_file, im_info, label_info):
        if im_info is None:
            im_info = dict()
        if isinstance(im_file, np.ndarray):
            if len(im_file.shape) != 3:
                raise Exception(
                    "im should be 3-dimensions, but now is {}-dimensions".
                    format(len(im_file.shape)))
            im = im_file
        else:
            try:
                im = cv2.imread(im_file).astype('float32')
            except:
                raise TypeError('Can\'t read The image file {}!'.format(
                    im_file))
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
        # make default im_info with [h, w, 1]
        im_info['im_resize_info'] = np.array(
            [im.shape[0], im.shape[1], 1.], dtype=np.float32)
        im_info['image_shape'] = np.array([im.shape[0],
                                           im.shape[1]]).astype('int32')
        use_mixup = False
        for t in transforms:
            if type(t).__name__ == 'MixupImage':
                use_mixup = True
            if not use_mixup:
                if 'mixup' in im_info:
                    del im_info['mixup']
        # decode mixup image
        if 'mixup' in im_info:
            im_info['mixup'] = \
              decode_image(im_info['mixup'][0],
                           im_info['mixup'][1],
                           im_info['mixup'][2])
        if label_info is None:
            return (im, im_info)
        else:
            return (im, im_info, label_info)

    outputs = decode_image(im, im_info, label_info)
    im = outputs[0]
    im_info = outputs[1]
    if len(outputs) == 3:
        label_info = outputs[2]
    if vdl_writer is not None:
        vdl_writer.add_image(
            tag='0. OriginalImage/' + str(step), img=im, step=0)
    op_id = 1
    bboxes = label_info['gt_bbox']
    transforms = [None] + transforms
    for op in transforms:
        if im is None:
            return None
        if isinstance(op, DetTransform) or op is None:
            if vdl_writer is not None and hasattr(op, 'prob'):
                op.prob = 1.0
            if op is not None:
                outputs = op(im, im_info, label_info)
            else:
                outputs = (im, im_info, label_info)
            im = outputs[0]
            vdl_im = im
            if vdl_writer is not None:
                if isinstance(op,
                              pdx.cv.transforms.det_transforms.ResizeByShort):
                    scale = outputs[1]['im_resize_info'][2]
                    bboxes = bboxes * scale
                elif isinstance(op, pdx.cv.transforms.det_transforms.Resize):
                    h = outputs[1]['image_shape'][0]
                    w = outputs[1]['image_shape'][1]
                    target_size = op.target_size
                    if isinstance(target_size, int):
                        h_scale = float(target_size) / h
                        w_scale = float(target_size) / w
                    else:
                        h_scale = float(target_size[0]) / h
                        w_scale = float(target_size[1]) / w
                    bboxes[:, 0] = bboxes[:, 0] * w_scale
                    bboxes[:, 1] = bboxes[:, 1] * h_scale
                    bboxes[:, 2] = bboxes[:, 2] * w_scale
                    bboxes[:, 3] = bboxes[:, 3] * h_scale
                else:
                    bboxes = outputs[2]['gt_bbox']
                if not isinstance(op, (
                        pdx.cv.transforms.det_transforms.RandomHorizontalFlip,
                        pdx.cv.transforms.det_transforms.Padding)):
                    for i in range(bboxes.shape[0]):
                        bbox = bboxes[i]
                        cname = labels[outputs[2]['gt_class'][i][0] - 1]
                        vdl_im = _draw_rectangle_and_cname(
                            vdl_im,
                            int(bbox[0]),
                            int(bbox[1]),
                            int(bbox[2]),
                            int(bbox[3]), cname,
                            catid2color[outputs[2]['gt_class'][i][0] - 1])
                if isinstance(op, pdx.cv.transforms.det_transforms.Normalize):
                    continue
        else:
            im = execute_imgaug(op, im)
            if label_info is not None:
                outputs = (im, im_info, label_info)
            else:
                outputs = (im, im_info)
            vdl_im = im
        if vdl_writer is not None:
            tag = str(op_id) + '. ' + op.__class__.__name__ + '/' + str(step)
            if op is None:
                tag = str(op_id) + '. OriginalImageWithGTBox/' + str(step)
            vdl_writer.add_image(tag=tag, img=vdl_im, step=0)
        op_id += 1


def seg_compose(im,
                im_info=None,
                label=None,
                transforms=None,
                vdl_writer=None,
                step=0):
    if im_info is None:
        im_info = list()
    if isinstance(im, np.ndarray):
        if len(im.shape) != 3:
            raise Exception(
                "im should be 3-dimensions, but now is {}-dimensions".format(
                    len(im.shape)))
    else:
        try:
            im = cv2.imread(im)
        except:
            raise ValueError('Can\'t read The image file {}!'.format(im))
    im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
    im = im.astype('float32')
    h, w, c = im.shape
    if label is not None:
        if not isinstance(label, np.ndarray):
            label = np.asarray(Image.open(label))
    if vdl_writer is not None:
        for i in range(0, c, 3):
            if c > 3:
                tag = '0. OriginalImage/{}_{}'.format(str(step), str(i // 3))
            else:
                tag = '0. OriginalImage/{}'.format(str(step))
            vdl_writer.add_image(tag=tag, img=im[:, :, i:i + 3], step=0)
    op_id = 1
    for op in transforms:
        if isinstance(op, SegTransform):
            outputs = op(im, im_info, label)
            im = outputs[0]
            if len(outputs) >= 2:
                im_info = outputs[1]
            if len(outputs) == 3:
                label = outputs[2]
            if isinstance(op, pdx.cv.transforms.seg_transforms.Normalize):
                continue
        else:
            im = execute_imgaug(op, im)
            if label is not None:
                outputs = (im, im_info, label)
            else:
                outputs = (im, im_info)
        if vdl_writer is not None:
            for i in range(0, c, 3):
                if c > 3:
                    tag = str(
                        op_id) + '. ' + op.__class__.__name__ + '/' + str(
                            step) + '_' + str(i // 3)
                else:
                    tag = str(
                        op_id) + '. ' + op.__class__.__name__ + '/' + str(step)
                vdl_writer.add_image(tag=tag, img=im[:, :, i:i + 3], step=0)
        op_id += 1


def visualize(dataset, img_count=3, save_dir='vdl_output'):
    '''对数据预处理/增强中间结果进行可视化。
    可使用VisualDL查看中间结果：
    1. VisualDL启动方式: visualdl --logdir vdl_output --port 8001
    2. 浏览器打开 https://0.0.0.0:8001即可，
        其中0.0.0.0为本机访问，如为远程服务, 改成相应机器IP

    Args:
        dataset (paddlex.datasets): 数据集读取器。
        img_count (int): 需要进行数据预处理/增强的图像数目。默认为3。
        save_dir (str): 日志保存的路径。默认为'vdl_output'。
    '''
    if dataset.num_samples < img_count:
        img_count = dataset.num_samples
    transforms = dataset.transforms
    if not osp.isdir(save_dir):
        if osp.exists(save_dir):
            os.remove(save_dir)
        os.makedirs(save_dir)
    from visualdl import LogWriter
    vdl_save_dir = osp.join(save_dir, 'image_transforms')
    vdl_writer = LogWriter(vdl_save_dir)
    for i, data in enumerate(dataset.iterator()):
        if i == img_count:
            break
        data.append(transforms.transforms)
        data.append(vdl_writer)
        data.append(i)
        if isinstance(transforms, ClsTransform):
            cls_compose(*data)
        elif isinstance(transforms, DetTransform):
            labels = dataset.labels
            color_map = get_color_map_list(len(labels) + 1)
            catid2color = {}
            for catid in range(len(labels)):
                catid2color[catid] = color_map[catid + 1]
            data.append(labels)
            data.append(catid2color)
            det_compose(*data)
        elif isinstance(transforms, SegTransform):
            seg_compose(*data)
        else:
            raise Exception('The transform must the subclass of \
                    ClsTransform or DetTransform or SegTransform!')



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import os
import shutil
import pickle
from os import path as osp
from .utils import set_folder_status, TaskStatus, copy_pretrained_model, PretrainedModelStatus
from . import workspace_pb2 as w


def list_pretrained_models(workspace):
    """列出预训练模型列表
    """
    pretrained_model_list = list()
    for id in workspace.pretrained_models:
        pretrained_model = workspace.pretrained_models[id]
        model_id = pretrained_model.id
        model_name = pretrained_model.name
        model_model = pretrained_model.model
        model_type = pretrained_model.type
        model_pid = pretrained_model.pid
        model_tid = pretrained_model.tid
        model_create_time = pretrained_model.create_time
        model_path = pretrained_model.path
        attr = {
            'id': model_id,
            'name': model_name,
            'model': model_model,
            'type': model_type,
            'pid': model_pid,
            'tid': model_tid,
            'create_time': model_create_time,
            'path': model_path
        }
        pretrained_model_list.append(attr)

    return {'status': 1, "pretrained_models": pretrained_model_list}


def create_pretrained_model(data, workspace, monitored_processes):
    """根据request创建预训练模型。

    Args:
        data为dict,key包括
        'pid'所属项目id, 'tid'所属任务id,'name'预训练模型名称,
        'source_path' 原模型路径, 'eval_results'（可选） 评估结果数据
    """
    time_array = time.localtime(time.time())
    create_time = time.strftime("%Y-%m-%d %H:%M:%S", time_array)
    id = workspace.max_pretrained_model_id + 1
    workspace.max_pretrained_model_id = id
    if id < 10000:
        id = 'PM%04d' % id
    else:
        id = 'PM{}'.format(id)
    pid = data['pid']
    tid = data['tid']
    name = data['name']
    source_path = data['source_path']
    assert pid in workspace.projects, "【预训练模型创建】项目ID'{}'不存在.".format(pid)
    assert tid in workspace.tasks, "【预训练模型创建】任务ID'{}'不存在.".format(tid)
    assert not id in workspace.pretrained_models, "【预训练模型创建】预训练模型'{}'已经被占用.".format(
        id)
    assert osp.exists(source_path), "原模型路径不存在: {}".format(source_path)
    path = osp.join(workspace.path, 'pretrain', id)
    if not osp.exists(path):
        os.makedirs(path)
    set_folder_status(path, PretrainedModelStatus.XPINIT)
    params = {'tid': tid}
    from .project.task import get_task_params
    ret = get_task_params(params, workspace)
    train_params = ret['train']
    model_structure = train_params.model
    if hasattr(train_params, "backbone"):
        model_structure = "{}-{}".format(model_structure,
                                         train_params.backbone)
    if hasattr(train_params, "with_fpn"):
        if train_params.with_fpn:
            model_structure = "{}-{}".format(model_structure, "WITH_FPN")

    pm = w.PretrainedModel(
        id=id,
        name=name,
        model=model_structure,
        type=workspace.projects[pid].type,
        pid=pid,
        tid=tid,
        create_time=create_time,
        path=path)
    workspace.pretrained_models[id].CopyFrom(pm)
    # 保存评估结果
    if 'eval_results' in data:
        with open(osp.join(source_path, "eval_res.pkl"), "wb") as f:
            pickle.dump(data['eval_results'], f)
    # 拷贝训练参数文件
    task_path = workspace.tasks[tid].path
    task_params_path = osp.join(task_path, 'params.pkl')
    if osp.exists(task_params_path):
        shutil.copy(task_params_path, path)
    # 拷贝数据集信息文件
    did = workspace.projects[pid].did
    dataset_path = workspace.datasets[did].path
    dataset_info_path = osp.join(dataset_path, "statis.pkl")
    if osp.exists(dataset_info_path):
        # 写入部分数据集信息
        with open(dataset_info_path, "rb") as f:
            dataset_info_dict = pickle.load(f)
        dataset_info_dict['name'] = workspace.datasets[did].name
        dataset_info_dict['desc'] = workspace.datasets[did].desc
        with open(dataset_info_path, "wb") as f:
            pickle.dump(dataset_info_dict, f)
        shutil.copy(dataset_info_path, path)

    # copy from source_path to path
    proc = copy_pretrained_model(source_path, path)
    monitored_processes.put(proc.pid)
    return {'status': 1, 'pmid': id}


def delete_pretrained_model(data, workspace):
    """删除pretrained_model。

    Args:
        data为dict,
        key包括'pmid'预训练模型id
    """
    pmid = data['pmid']
    assert pmid in workspace.pretrained_models, "预训练模型ID'{}'不存在.".format(pmid)
    if osp.exists(workspace.pretrained_models[pmid].path):
        shutil.rmtree(workspace.pretrained_models[pmid].path)
    del workspace.pretrained_models[pmid]
    return {'status': 1}


def create_exported_model(data, workspace):
    """根据request创建已发布模型。
    Args:
        data为dict,key包括
        'pid'所属项目id, 'tid'所属任务id,'name'已发布模型名称,
        'path' 模型路径, 'exported_type' 已发布模型类型,
    """
    time_array = time.localtime(time.time())
    create_time = time.strftime("%Y-%m-%d %H:%M:%S", time_array)
    emid = workspace.max_exported_model_id + 1
    workspace.max_exported_model_id = emid
    if emid < 10000:
        emid = 'EM%04d' % emid
    else:
        emid = 'EM{}'.format(emid)
    pid = data['pid']
    tid = data['tid']
    name = data['name']
    path = data['path']
    exported_type = data['exported_type']
    assert pid in workspace.projects, "【已发布模型创建】项目ID'{}'不存在.".format(pid)
    assert tid in workspace.tasks, "【已发布模型创建】任务ID'{}'不存在.".format(tid)
    assert emid not in workspace.exported_models, "【已发布模型创建】已发布模型'{}'已经被占用.".format(
        emid)
    #assert osp.exists(path), "已发布模型路径不存在: {}".format(path)
    if not osp.exists(path):
        os.makedirs(path)
    task_path = workspace.tasks[tid].path
    # 拷贝评估结果
    eval_res_path = osp.join(task_path, 'eval_res.pkl')
    if osp.exists(eval_res_path):
        shutil.copy(eval_res_path, path)
    # 拷贝训练参数文件
    task_params_path = osp.join(task_path, 'params.pkl')
    if osp.exists(task_params_path):
        shutil.copy(task_params_path, path)
    # 拷贝数据集信息文件
    did = workspace.projects[pid].did
    dataset_path = workspace.datasets[did].path
    dataset_info_path = osp.join(dataset_path, "statis.pkl")
    if osp.exists(dataset_info_path):
        # 写入部分数据集信息
        with open(dataset_info_path, "rb") as f:
            dataset_info_dict = pickle.load(f)
        dataset_info_dict['name'] = workspace.datasets[did].name
        dataset_info_dict['desc'] = workspace.datasets[did].desc
        with open(dataset_info_path, "wb") as f:
            pickle.dump(dataset_info_dict, f)
        shutil.copy(dataset_info_path, path)
    from .project.task import get_task_params
    params = {'tid': tid}
    ret = get_task_params(params, workspace)
    train_params = ret['train']
    model_structure = train_params.model
    if hasattr(train_params, "backbone"):
        model_structure = "{}-{}".format(model_structure,
                                         train_params.backbone)
    if hasattr(train_params, "with_fpn"):
        if train_params.with_fpn:
            model_structure = "{}-{}".format(model_structure, "WITH_FPN")

    em = w.ExportedModel(
        id=emid,
        name=name,
        model=model_structure,
        type=workspace.projects[pid].type,
        pid=pid,
        tid=tid,
        create_time=create_time,
        path=path,
        exported_type=exported_type)

    workspace.exported_models[emid].CopyFrom(em)
    return {'status': 1, 'emid': emid}


def list_exported_models(workspace):
    """列出预训练模型列表，可根据request中的参数进行筛选

    Args:
    """
    exported_model_list = list()
    for id in workspace.exported_models:
        exported_model = workspace.exported_models[id]
        model_id = exported_model.id
        model_name = exported_model.name
        model_model = exported_model.model
        model_type = exported_model.type
        model_pid = exported_model.pid
        model_tid = exported_model.tid
        model_create_time = exported_model.create_time
        model_path = exported_model.path
        model_exported_type = exported_model.exported_type
        attr = {
            'id': model_id,
            'name': model_name,
            'model': model_model,
            'type': model_type,
            'pid': model_pid,
            'tid': model_tid,
            'create_time': model_create_time,
            'path': model_path,
            'exported_type': model_exported_type
        }
        if model_tid in workspace.tasks:
            from .project.task import get_export_status
            params = {'tid': model_tid}
            results = get_export_status(params, workspace)
            if results['export_status'] == TaskStatus.XEXPORTED:
                exported_model_list.append(attr)
        else:
            exported_model_list.append(attr)
    return {'status': 1, "exported_models": exported_model_list}


def delete_exported_model(data, workspace):
    """删除exported_model。

    Args:
        data为dict,
        key包括'emid'已发布模型id
    """
    emid = data['emid']
    assert emid in workspace.exported_models, "已发布模型模型ID'{}'不存在.".format(emid)
    if osp.exists(workspace.exported_models[emid].path):
        shutil.rmtree(workspace.exported_models[emid].path)
    del workspace.exported_models[emid]
    return {'status': 1}


def get_model_details(data, workspace):
    """获取模型详情。

    Args:
        data为dict,
        key包括'mid'模型id
    """
    mid = data['mid']
    if mid in workspace.pretrained_models:
        model_path = workspace.pretrained_models[mid].path
    elif mid in workspace.exported_models:
        model_path = workspace.exported_models[mid].path
    else:
        raise "模型{}不存在".format(mid)
    dataset_file = osp.join(model_path, 'statis.pkl')
    dataset_info = pickle.load(open(dataset_file, 'rb'))
    dataset_attr = {
        'name': dataset_info['name'],
        'desc': dataset_info['desc'],
        'labels': dataset_info['labels'],
        'train_num': len(dataset_info['train_files']),
        'val_num': len(dataset_info['val_files']),
        'test_num': len(dataset_info['test_files'])
    }
    task_params_file = osp.join(model_path, 'params.pkl')
    task_params = pickle.load(open(task_params_file, 'rb'))
    eval_result_file = osp.join(model_path, 'eval_res.pkl')
    eval_result = pickle.load(open(eval_result_file, 'rb'))
    #model_file = {'task_attr': task_params_file, 'eval_result': eval_result_file}
    return {
        'status': 1,
        'dataset_attr': dataset_attr,
        'task_params': task_params,
        'eval_result': eval_result
    }



# copytrue (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path as osp
import os
import numpy as np
from PIL import Image
import sys
import cv2
import psutil
import shutil
import pickle
import base64
import multiprocessing as mp
from ..utils import (pkill, set_folder_status, get_folder_status, TaskStatus,
                     PredictStatus, PruneStatus)
from .evaluate.draw_pred_result import visualize_classified_result, visualize_detected_result, visualize_segmented_result
from .visualize import plot_det_label, plot_insseg_label, get_color_map_list


def _call_paddle_prune(best_model_path, prune_analysis_path, params):
    mode = 'w'
    sys.stdout = open(
        osp.join(prune_analysis_path, 'out.log'), mode, encoding='utf-8')
    sys.stderr = open(
        osp.join(prune_analysis_path, 'err.log'), mode, encoding='utf-8')
    sensitivities_path = osp.join(prune_analysis_path, "sensitivities.data")
    task_type = params['task_type']
    dataset_path = params['dataset_path']
    os.environ['CUDA_VISIBLE_DEVICES'] = params['train'].cuda_visible_devices
    if task_type == "classification":
        from .prune.classification import prune
    elif task_type in ["detection", "instance_segmentation"]:
        from .prune.detection import prune
    elif task_type == "segmentation":
        from .prune.segmentation import prune
    batch_size = params['train'].batch_size
    prune(best_model_path, dataset_path, sensitivities_path, batch_size)
    import paddlex as pdx
    from paddlex.cv.models.slim.visualize import visualize
    model = pdx.load_model(best_model_path)
    visualize(model, sensitivities_path, prune_analysis_path)
    set_folder_status(prune_analysis_path, PruneStatus.XSPRUNEDONE)


def _call_paddlex_train(task_path, params):
    '''
    Args:
        params为dict，字段包括'pretrain_weights_download_save_dir': 预训练模型保存路径，
        'task_type': 任务类型，'dataset_path': 数据集路径，'train':训练参数
    '''

    mode = 'w'
    if params['train'].resume_checkpoint is not None:
        mode = 'a'
    sys.stdout = open(osp.join(task_path, 'out.log'), mode, encoding='utf-8')
    sys.stderr = open(osp.join(task_path, 'err.log'), mode, encoding='utf-8')
    sys.stdout.write("This log file path is {}\n".format(
        osp.join(task_path, 'out.log')))
    sys.stdout.write("注意：标志为WARNING/INFO类的仅为警告或提示类信息，非错误信息\n")
    sys.stderr.write("This log file path is {}\n".format(
        osp.join(task_path, 'err.log')))
    sys.stderr.write("注意：标志为WARNING/INFO类的仅为警告或提示类信息，非错误信息\n")
    os.environ['CUDA_VISIBLE_DEVICES'] = params['train'].cuda_visible_devices
    import paddlex as pdx
    pdx.gui_mode = True
    pdx.log_level = 3
    pdx.pretrain_dir = params['pretrain_weights_download_save_dir']
    task_type = params['task_type']
    dataset_path = params['dataset_path']
    if task_type == "classification":
        from .train.classification import train
    elif task_type in ["detection", "instance_segmentation"]:
        from .train.detection import train
    elif task_type == "segmentation":
        from .train.segmentation import train
    train(task_path, dataset_path, params['train'])
    set_folder_status(task_path, TaskStatus.XTRAINDONE)


def _call_paddlex_evaluate_model(task_path,
                                 model_path,
                                 task_type,
                                 epoch,
                                 topk=5,
                                 score_thresh=0.3,
                                 overlap_thresh=0.5):
    evaluate_status_path = osp.join(task_path, './logs/evaluate')
    sys.stdout = open(
        osp.join(evaluate_status_path, 'out.log'), 'w', encoding='utf-8')
    sys.stderr = open(
        osp.join(evaluate_status_path, 'err.log'), 'w', encoding='utf-8')
    if task_type == "classification":
        from .evaluate.classification import Evaluator
        evaluator = Evaluator(model_path, topk=topk)
    elif task_type == "detection":
        from .evaluate.detection import DetEvaluator
        evaluator = DetEvaluator(
            model_path,
            score_threshold=score_thresh,
            overlap_thresh=overlap_thresh)
    elif task_type == "instance_segmentation":
        from .evaluate.detection import InsSegEvaluator
        evaluator = InsSegEvaluator(
            model_path,
            score_threshold=score_thresh,
            overlap_thresh=overlap_thresh)
    elif task_type == "segmentation":
        from .evaluate.segmentation import Evaluator
        evaluator = Evaluator(model_path)
    report = evaluator.generate_report()
    report['epoch'] = epoch
    pickle.dump(report, open(osp.join(task_path, "eval_res.pkl"), "wb"))
    set_folder_status(evaluate_status_path, TaskStatus.XEVALUATED)
    set_folder_status(task_path, TaskStatus.XEVALUATED)


def _call_paddlex_predict(task_path,
                          predict_status_path,
                          params,
                          img_list,
                          img_data,
                          save_dir,
                          score_thresh,
                          epoch=None):
    total_num = open(
        osp.join(predict_status_path, 'total_num'), 'w', encoding='utf-8')

    def write_file_num(total_file_num):
        total_num.write(str(total_file_num))
        total_num.close()

    sys.stdout = open(
        osp.join(predict_status_path, 'out.log'), 'w', encoding='utf-8')
    sys.stderr = open(
        osp.join(predict_status_path, 'err.log'), 'w', encoding='utf-8')

    import paddlex as pdx
    pdx.log_level = 3
    task_type = params['task_type']
    dataset_path = params['dataset_path']
    if epoch is None:
        model_path = osp.join(task_path, 'output', 'best_model')
    else:
        model_path = osp.join(task_path, 'output', 'epoch_{}'.format(epoch))
    model = pdx.load_model(model_path)
    file_list = dict()
    predicted_num = 0
    if task_type == "classification":
        if img_data is None:
            if len(img_list) == 0 and osp.exists(
                    osp.join(dataset_path, "test_list.txt")):
                with open(osp.join(dataset_path, "test_list.txt")) as f:
                    for line in f:
                        items = line.strip().split()
                        file_list[osp.join(dataset_path, items[0])] = items[1]
            else:
                for image in img_list:
                    file_list[image] = None
            total_file_num = len(file_list)
            write_file_num(total_file_num)
            for image, label_id in file_list.items():
                pred_result = {}
                if label_id is not None:
                    pred_result["gt_label"] = model.labels[int(label_id)]
                results = model.predict(img_file=image)
                pred_result["label"] = []
                pred_result["score"] = []
                pred_result["topk"] = len(results)
                for res in results:
                    pred_result["label"].append(res['category'])
                    pred_result["score"].append(res['score'])
                visualize_classified_result(save_dir, image, pred_result)
                predicted_num += 1
        else:
            img_data = base64.b64decode(img_data)
            img_array = np.frombuffer(img_data, np.uint8)
            img = cv2.imdecode(img_array, cv2.COLOR_RGB2BGR)
            results = model.predict(img)
            pred_result = {}
            pred_result["label"] = []
            pred_result["score"] = []
            pred_result["topk"] = len(results)
            for res in results:
                pred_result["label"].append(res['category'])
                pred_result["score"].append(res['score'])
            visualize_classified_result(save_dir, img, pred_result)
    elif task_type in ["detection", "instance_segmentation"]:
        if img_data is None:
            if task_type == "detection" and osp.exists(
                    osp.join(dataset_path, "test_list.txt")):
                if len(img_list) == 0 and osp.exists(
                        osp.join(dataset_path, "test_list.txt")):
                    with open(osp.join(dataset_path, "test_list.txt")) as f:
                        for line in f:
                            items = line.strip().split()
                            file_list[osp.join(dataset_path, items[0])] = \
                                osp.join(dataset_path, items[1])
                else:
                    for image in img_list:
                        file_list[image] = None
                total_file_num = len(file_list)
                write_file_num(total_file_num)
                for image, anno in file_list.items():
                    results = model.predict(img_file=image)
                    image_pred = pdx.det.visualize(
                        image, results, threshold=score_thresh, save_dir=None)
                    save_name = osp.join(save_dir, osp.split(image)[-1])
                    image_gt = None
                    if anno is not None:
                        image_gt = plot_det_label(image, anno, model.labels)
                    visualize_detected_result(save_name, image_gt, image_pred)
                    predicted_num += 1
            elif len(img_list) == 0 and osp.exists(
                    osp.join(dataset_path, "test.json")):
                from pycocotools.coco import COCO
                anno_path = osp.join(dataset_path, "test.json")
                coco = COCO(anno_path)
                img_ids = coco.getImgIds()
                total_file_num = len(img_ids)
                write_file_num(total_file_num)
                for img_id in img_ids:
                    img_anno = coco.loadImgs(img_id)[0]
                    file_name = img_anno['file_name']
                    name = (osp.split(file_name)[-1]).split(".")[0]
                    anno = osp.join(dataset_path, "Annotations", name + ".npy")
                    img_file = osp.join(dataset_path, "JPEGImages", file_name)
                    results = model.predict(img_file=img_file)
                    image_pred = pdx.det.visualize(
                        img_file,
                        results,
                        threshold=score_thresh,
                        save_dir=None)
                    save_name = osp.join(save_dir, osp.split(img_file)[-1])
                    if task_type == "detection":
                        image_gt = plot_det_label(img_file, anno, model.labels)
                    else:
                        image_gt = plot_insseg_label(img_file, anno,
                                                     model.labels)
                    visualize_detected_result(save_name, image_gt, image_pred)
                    predicted_num += 1
            else:
                total_file_num = len(img_list)
                write_file_num(total_file_num)
                for image in img_list:
                    results = model.predict(img_file=image)
                    image_pred = pdx.det.visualize(
                        image, results, threshold=score_thresh, save_dir=None)
                    save_name = osp.join(save_dir, osp.split(image)[-1])
                    visualize_detected_result(save_name, None, image_pred)
                    predicted_num += 1
        else:
            img_data = base64.b64decode(img_data)
            img_array = np.frombuffer(img_data, np.uint8)
            img = cv2.imdecode(img_array, cv2.COLOR_RGB2BGR)
            results = model.predict(img)
            image_pred = pdx.det.visualize(
                img, results, threshold=score_thresh, save_dir=None)
            image_gt = None
            save_name = osp.join(save_dir, 'predict_result.png')
            visualize_detected_result(save_name, image_gt, image_pred)

    elif task_type == "segmentation":
        if img_data is None:
            if len(img_list) == 0 and osp.exists(
                    osp.join(dataset_path, "test_list.txt")):
                with open(osp.join(dataset_path, "test_list.txt")) as f:
                    for line in f:
                        items = line.strip().split()
                        file_list[osp.join(dataset_path, items[0])] = \
                            osp.join(dataset_path, items[1])
            else:
                for image in img_list:
                    file_list[image] = None
            total_file_num = len(file_list)
            write_file_num(total_file_num)
            color_map = get_color_map_list(256)
            legend = {}
            for i in range(len(model.labels)):
                legend[model.labels[i]] = color_map[i]
            for image, anno in file_list.items():
                results = model.predict(img_file=image)
                image_pred = pdx.seg.visualize(image, results, save_dir=None)
                pse_pred = pdx.seg.visualize(
                    image, results, weight=0, save_dir=None)
                image_ground = None
                pse_label = None
                if anno is not None:
                    label = np.asarray(Image.open(anno)).astype('uint8')
                    image_ground = pdx.seg.visualize(
                        image, {'label_map': label}, save_dir=None)
                    pse_label = pdx.seg.visualize(
                        image, {'label_map': label}, weight=0, save_dir=None)
                save_name = osp.join(save_dir, osp.split(image)[-1])
                visualize_segmented_result(save_name, image_ground, pse_label,
                                           image_pred, pse_pred, legend)
                predicted_num += 1
        else:
            img_data = base64.b64decode(img_data)
            img_array = np.frombuffer(img_data, np.uint8)
            img = cv2.imdecode(img_array, cv2.COLOR_RGB2BGR)
            color_map = get_color_map_list(256)
            legend = {}
            for i in range(len(model.labels)):
                legend[model.labels[i]] = color_map[i]
            results = model.predict(img)
            image_pred = pdx.seg.visualize(image, results, save_dir=None)
            pse_pred = pdx.seg.visualize(
                image, results, weight=0, save_dir=None)
            image_ground = None
            pse_label = None
            save_name = osp.join(save_dir, 'predict_result.png')
            visualize_segmented_result(save_name, image_ground, pse_label,
                                       image_pred, pse_pred, legend)
    set_folder_status(predict_status_path, PredictStatus.XPREDONE)


def _call_paddlex_export_infer(task_path, save_dir, export_status_path, epoch):
    # 导出模型不使用GPU
    sys.stdout = open(
        osp.join(export_status_path, 'out.log'), 'w', encoding='utf-8')
    sys.stderr = open(
        osp.join(export_status_path, 'err.log'), 'w', encoding='utf-8')
    import os
    os.environ['CUDA_VISIBLE_DEVICES'] = ''
    import paddlex as pdx
    if epoch is not None:
        model_dir = "epoch_{}".format(epoch)
        model_path = osp.join(task_path, 'output', model_dir)
    else:
        model_path = osp.join(task_path, 'output', 'best_model')
    model = pdx.load_model(model_path)
    model.export_inference_model(save_dir)
    set_folder_status(export_status_path, TaskStatus.XEXPORTED)
    set_folder_status(task_path, TaskStatus.XEXPORTED)


def _call_paddlex_export_quant(task_path, params, save_dir, export_status_path,
                               epoch):
    sys.stdout = open(
        osp.join(export_status_path, 'out.log'), 'w', encoding='utf-8')
    sys.stderr = open(
        osp.join(export_status_path, 'err.log'), 'w', encoding='utf-8')
    dataset_path = params['dataset_path']
    task_type = params['task_type']
    os.environ['CUDA_VISIBLE_DEVICES'] = params['train'].cuda_visible_devices
    import paddlex as pdx
    if epoch is not None:
        model_dir = "epoch_{}".format(epoch)
        model_path = osp.join(task_path, 'output', model_dir)
    else:
        model_path = osp.join(task_path, 'output', 'best_model')
    model = pdx.load_model(model_path)
    if task_type == "classification":
        train_file_list = osp.join(dataset_path, 'train_list.txt')
        val_file_list = osp.join(dataset_path, 'val_list.txt')
        label_list = osp.join(dataset_path, 'labels.txt')
        quant_dataset = pdx.datasets.ImageNet(
            data_dir=dataset_path,
            file_list=train_file_list,
            label_list=label_list,
            transforms=model.test_transforms)
        eval_dataset = pdx.datasets.ImageNet(
            data_dir=dataset_path,
            file_list=val_file_list,
            label_list=label_list,
            transforms=model.eval_transforms)
    elif task_type == "detection":
        train_file_list = osp.join(dataset_path, 'train_list.txt')
        val_file_list = osp.join(dataset_path, 'val_list.txt')
        label_list = osp.join(dataset_path, 'labels.txt')
        quant_dataset = pdx.datasets.VOCDetection(
            data_dir=dataset_path,
            file_list=train_file_list,
            label_list=label_list,
            transforms=model.test_transforms)
        eval_dataset = pdx.datasets.VOCDetection(
            data_dir=dataset_path,
            file_list=val_file_list,
            label_list=label_list,
            transforms=model.eval_transforms)
    elif task_type == "instance_segmentation":
        train_json = osp.join(dataset_path, 'train.json')
        val_json = osp.join(dataset_path, 'val.json')
        quant_dataset = pdx.datasets.CocoDetection(
            data_dir=osp.join(dataset_path, 'JPEGImages'),
            ann_file=train_json,
            transforms=model.test_transforms)
        eval_dataset = pdx.datasets.CocoDetection(
            data_dir=osp.join(dataset_path, 'JPEGImages'),
            ann_file=val_json,
            transforms=model.eval_transforms)
    elif task_type == "segmentation":
        train_file_list = osp.join(dataset_path, 'train_list.txt')
        val_file_list = osp.join(dataset_path, 'val_list.txt')
        label_list = osp.join(dataset_path, 'labels.txt')
        quant_dataset = pdx.datasets.SegDataset(
            data_dir=dataset_path,
            file_list=train_file_list,
            label_list=label_list,
            transforms=model.test_transforms)
        eval_dataset = pdx.datasets.SegDataset(
            data_dir=dataset_path,
            file_list=val_file_list,
            label_list=label_list,
            transforms=model.eval_transforms)
    metric_before = model.evaluate(eval_dataset)
    pdx.log_level = 3
    pdx.slim.export_quant_model(
        model, quant_dataset, batch_size=1, save_dir=save_dir, cache_dir=None)
    model_quant = pdx.load_model(save_dir)
    metric_after = model_quant.evaluate(eval_dataset)
    metrics = {}
    if task_type == "segmentation":
        metrics['before'] = {'miou': metric_before['miou']}
        metrics['after'] = {'miou': metric_after['miou']}
    else:
        metrics['before'] = metric_before
        metrics['after'] = metric_after
    import json
    with open(
            osp.join(export_status_path, 'quant_result.json'),
            'w',
            encoding='utf-8') as f:
        json.dump(metrics, f)
    set_folder_status(export_status_path, TaskStatus.XEXPORTED)
    set_folder_status(task_path, TaskStatus.XEXPORTED)


def _call_paddlelite_export_lite(model_path, save_dir=None, place="arm"):
    import paddlelite.lite as lite
    opt = lite.Opt()
    model_file = os.path.join(model_path, '__model__')
    params_file = os.path.join(model_path, '__params__')
    if save_dir is None:
        save_dir = osp.join(model_path, "lite_model")
    if not osp.exists(save_dir):
        os.makedirs(save_dir)
    path = osp.join(save_dir, "model")
    opt.run_optimize("", model_file, params_file, "naive_buffer", place, path)


def safe_clean_folder(folder):
    if osp.exists(folder):
        try:
            shutil.rmtree(folder)
            os.makedirs(folder)
        except Exception as e:
            pass
        if osp.exists(folder):
            for root, dirs, files in os.walk(folder):
                for name in files:
                    try:
                        os.remove(os.path.join(root, name))
                    except Exception as e:
                        pass
        else:
            os.makedirs(folder)
    else:
        os.makedirs(folder)
    if not osp.exists(folder):
        os.makedirs(folder)


def get_task_max_saved_epochs(task_path):
    saved_epoch_num = -1
    output_path = osp.join(task_path, "output")
    if osp.exists(output_path):
        for f in os.listdir(output_path):
            if f.startswith("epoch_"):
                if not osp.exists(osp.join(output_path, f, '.success')):
                    continue
                curr_epoch_num = int(f[6:])
                if curr_epoch_num > saved_epoch_num:
                    saved_epoch_num = curr_epoch_num
    return saved_epoch_num


def get_task_status(task_path):
    status, message = get_folder_status(task_path, True)
    task_id = os.path.split(task_path)[-1]
    err_log = os.path.join(task_path, 'err.log')
    if status in [TaskStatus.XTRAINING, TaskStatus.XPRUNETRAIN]:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = TaskStatus.XTRAINFAIL
            message = "训练任务{}异常终止，请查阅错误日志具体确认原因{}。\n\n 如若通过日志无法确定原因，可尝试以下几种方法，\n" \
            "1. 尝试重新启动训练，看是否能正常训练; \n" \
            "2. 调低batch_size（需同时按比例调低学习率等参数）排除是否是显存或内存不足的原因导致;\n" \
            "3. 前往GitHub提ISSUE，描述清楚问题会有工程师及时回复： https://github.com/PaddlePaddle/PaddleX/issues ; \n" \
            "3. 加QQ群1045148026或邮件至paddlex@baidu.com在线咨询工程师".format(task_id, err_log)
            set_folder_status(task_path, status, message)
    return status, message


def train_model(task_path):
    """训练模型

    Args:
        task_path(str): 模型训练的参数保存在task_path下的'params.pkl'文件中
    """
    params_conf_file = osp.join(task_path, 'params.pkl')
    assert osp.exists(
        params_conf_file), "任务无法启动，路径{}下不存在参数配置文件params.pkl".format(task_path)
    with open(params_conf_file, 'rb') as f:
        params = pickle.load(f)
    sensitivities_path = params['train'].sensitivities_path
    p = mp.Process(target=_call_paddlex_train, args=(task_path, params))
    p.start()
    if sensitivities_path is None:
        set_folder_status(task_path, TaskStatus.XTRAINING, p.pid)
    else:
        set_folder_status(task_path, TaskStatus.XPRUNETRAIN, p.pid)
    return p


def stop_train_model(task_path):
    """停止正在训练的模型

    Args:
        task_path(str): 从task_path下的'XTRANING'文件中获取训练的进程id
    """
    status, message = get_task_status(task_path)
    if status in [TaskStatus.XTRAINING, TaskStatus.XPRUNETRAIN]:
        pid = int(message)
        pkill(pid)
        best_model_saved = True
        if not osp.exists(osp.join(task_path, 'output', 'best_model')):
            best_model_saved = False
        set_folder_status(task_path, TaskStatus.XTRAINEXIT, best_model_saved)
    else:
        raise Exception("模型训练任务没在运行中")


def prune_analysis_model(task_path):
    """模型裁剪分析

    Args:
        task_path(str): 模型训练的参数保存在task_path
        dataset_path(str) 模型裁剪中评估数据集的路径
    """
    best_model_path = osp.join(task_path, 'output', 'best_model')
    assert osp.exists(best_model_path), "该任务暂未保存模型，无法进行模型裁剪分析"
    prune_analysis_path = osp.join(task_path, 'prune')
    if not osp.exists(prune_analysis_path):
        os.makedirs(prune_analysis_path)

    params_conf_file = osp.join(task_path, 'params.pkl')
    assert osp.exists(
        params_conf_file), "任务无法启动，路径{}下不存在参数配置文件params.pkl".format(task_path)
    with open(params_conf_file, 'rb') as f:
        params = pickle.load(f)
    assert params['train'].model.lower() not in [
        "ppyolo", "fasterrcnn", "maskrcnn", "fastscnn", "HRNet_W18"
    ], "暂不支持PPYOLO、FasterRCNN、MaskRCNN、HRNet_W18、FastSCNN模型裁剪"
    p = mp.Process(
        target=_call_paddle_prune,
        args=(best_model_path, prune_analysis_path, params))
    p.start()
    set_folder_status(prune_analysis_path, PruneStatus.XSPRUNEING, p.pid)
    set_folder_status(task_path, TaskStatus.XPRUNEING, p.pid)
    return p


def get_prune_status(prune_path):
    status, message = get_folder_status(prune_path, True)
    if status in [PruneStatus.XSPRUNEING]:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = PruneStatus.XSPRUNEFAIL
            message = "模型裁剪异常终止，可能原因如下：\n1.暂不支持FasterRCNN、MaskRCNN模型的模型裁剪\n2.模型裁剪过程中进程被异常结束，建议重新启动模型裁剪任务"
            set_folder_status(prune_path, status, message)
    return status, message


def stop_prune_analysis(prune_path):
    """停止正在裁剪分析的模型

    Args:
        prune_path(str): prune_path'XSSLMING'文件中获取训练的进程id
    """
    status, message = get_prune_status(prune_path)
    if status == PruneStatus.XSPRUNEING:
        pid = int(message)
        pkill(pid)
        set_folder_status(prune_path, PruneStatus.XSPRUNEEXIT)
    else:
        raise Exception("模型裁剪分析任务未在运行中")


def evaluate_model(task_path,
                   task_type,
                   epoch=None,
                   topk=5,
                   score_thresh=0.3,
                   overlap_thresh=0.5):
    """评估最优模型

    Args:
        task_path(str): 模型训练相关结果的保存路径
    """
    output_path = osp.join(task_path, 'output')
    if not osp.exists(osp.join(output_path, 'best_model')):
        raise Exception("未在训练路径{}下发现保存的best_model，无法进行评估".format(output_path))
    evaluate_status_path = osp.join(task_path, './logs/evaluate')
    safe_clean_folder(evaluate_status_path)
    if epoch is None:
        model_path = osp.join(output_path, 'best_model')
    else:
        epoch_dir = "{}_{}".format('epoch', epoch)
        model_path = osp.join(output_path, epoch_dir)
    p = mp.Process(
        target=_call_paddlex_evaluate_model,
        args=(task_path, model_path, task_type, epoch, topk, score_thresh,
              overlap_thresh))
    p.start()
    set_folder_status(evaluate_status_path, TaskStatus.XEVALUATING, p.pid)
    return p


def get_evaluate_status(task_path):
    """获取导出状态
    Args:
        task_path(str): 训练任务文件夹
    """
    evaluate_status_path = osp.join(task_path, './logs/evaluate')
    if not osp.exists(evaluate_status_path):
        return None, "No evaluate fold in path {}".format(task_path)
    status, message = get_folder_status(evaluate_status_path, True)
    if status == TaskStatus.XEVALUATING:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = TaskStatus.XEVALUATEFAIL
            message = "评估过程出现异常，请尝试重新评估！"
            set_folder_status(evaluate_status_path, status, message)
    if status not in [
            TaskStatus.XEVALUATING, TaskStatus.XEVALUATED,
            TaskStatus.XEVALUATEFAIL
    ]:
        raise ValueError("Wrong status in evaluate task {}".format(status))
    return status, message


def get_predict_status(task_path):
    """获取预测任务状态

    Args:
        task_path(str): 从predict_path下的'XPRESTART'文件中获取训练的进程id
    """
    from ..utils import list_files
    predict_status_path = osp.join(task_path, "./logs/predict")
    save_dir = osp.join(task_path, "visualized_test_results")
    if not osp.exists(save_dir):
        return None, "任务目录下没有visualized_test_results文件夹，{}".format(
            task_path), 0, 0
    status, message = get_folder_status(predict_status_path, True)
    if status == PredictStatus.XPRESTART:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = PredictStatus.XPREFAIL
            message = "图片预测过程出现异常，请尝试重新预测！"
            set_folder_status(predict_status_path, status, message)
    if status not in [
            PredictStatus.XPRESTART, PredictStatus.XPREDONE,
            PredictStatus.XPREFAIL
    ]:
        raise ValueError("预测任务状态异常，{}".format(status))
    predict_num = len(list_files(save_dir))
    if predict_num > 0:
        if predict_num == 1:
            total_num = 1
        else:
            total_num = int(
                open(
                    osp.join(predict_status_path, "total_num"),
                    encoding='utf-8').readline().strip())
    else:
        predict_num = 0
        total_num = 0
    return status, message, predict_num, total_num


def predict_test_pics(task_path,
                      img_list=[],
                      img_data=None,
                      save_dir=None,
                      score_thresh=0.5,
                      epoch=None):
    """模型预测

    Args:
        task_path(str): 模型训练的参数保存在task_path下的'params.pkl'文件中
    """
    params_conf_file = osp.join(task_path, 'params.pkl')
    assert osp.exists(
        params_conf_file), "任务无法启动，路径{}下不存在参数配置文件params.pkl".format(task_path)
    with open(params_conf_file, 'rb') as f:
        params = pickle.load(f)
    predict_status_path = osp.join(task_path, "./logs/predict")
    safe_clean_folder(predict_status_path)
    save_dir = osp.join(task_path, 'visualized_test_results')
    safe_clean_folder(save_dir)
    p = mp.Process(
        target=_call_paddlex_predict,
        args=(task_path, predict_status_path, params, img_list, img_data,
              save_dir, score_thresh, epoch))
    p.start()
    set_folder_status(predict_status_path, PredictStatus.XPRESTART, p.pid)
    return p, save_dir


def stop_predict_task(task_path):
    """停止预测任务

    Args:
        task_path(str): 从predict_path下的'XPRESTART'文件中获取训练的进程id
    """
    from ..utils import list_files
    predict_status_path = osp.join(task_path, "./logs/predict")
    save_dir = osp.join(task_path, "visualized_test_results")
    if not osp.exists(save_dir):
        return None, "任务目录下没有visualized_test_results文件夹，{}".format(
            task_path), 0, 0
    status, message = get_folder_status(predict_status_path, True)
    if status == PredictStatus.XPRESTART:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = PredictStatus.XPREFAIL
            message = "图片预测过程出现异常，请尝试重新预测！"
            set_folder_status(predict_status_path, status, message)
        else:
            pkill(pid)
            status = PredictStatus.XPREFAIL
            message = "图片预测进程已停止！"
            set_folder_status(predict_status_path, status, message)
    if status not in [
            PredictStatus.XPRESTART, PredictStatus.XPREDONE,
            PredictStatus.XPREFAIL
    ]:
        raise ValueError("预测任务状态异常，{}".format(status))
    predict_num = len(list_files(save_dir))
    if predict_num > 0:
        total_num = int(
            open(
                osp.join(predict_status_path, "total_num"), encoding='utf-8')
            .readline().strip())
    else:
        predict_num = 0
        total_num = 0
    return status, message, predict_num, total_num


def get_export_status(task_path):
    """获取导出状态

    Args:
        task_path(str): 从task_path下的'export/XEXPORTING'文件中获取训练的进程id
    Return:
        导出的状态和其他消息.
    """
    export_status_path = osp.join(task_path, './logs/export')
    if not osp.exists(export_status_path):
        return None, "{}任务目录下没有export文件夹".format(task_path)
    status, message = get_folder_status(export_status_path, True)
    if status == TaskStatus.XEXPORTING:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = TaskStatus.XEXPORTFAIL
            message = "导出过程出现异常，请尝试重新评估！"
            set_folder_status(export_status_path, status, message)
    if status not in [
            TaskStatus.XEXPORTING, TaskStatus.XEXPORTED, TaskStatus.XEXPORTFAIL
    ]:
        # raise ValueError("获取到的导出状态异常，{}。".format(status))
        return None, "获取到的导出状态异常，{}。".format(status)
    return status, message


def export_quant_model(task_path, save_dir, epoch=None):
    """导出量化模型

    Args:
        task_path(str): 模型训练的路径
        save_dir(str): 导出后的模型保存路径
    """
    output_path = osp.join(task_path, 'output')
    if not osp.exists(osp.join(output_path, 'best_model')):
        raise Exception("未在训练路径{}下发现保存的best_model，导出失败".format(output_path))
    export_status_path = osp.join(task_path, './logs/export')
    safe_clean_folder(export_status_path)

    params_conf_file = osp.join(task_path, 'params.pkl')
    assert osp.exists(
        params_conf_file), "任务无法启动，路径{}下不存在参数配置文件params.pkl".format(task_path)
    with open(params_conf_file, 'rb') as f:
        params = pickle.load(f)
    p = mp.Process(
        target=_call_paddlex_export_quant,
        args=(task_path, params, save_dir, export_status_path, epoch))
    p.start()
    set_folder_status(export_status_path, TaskStatus.XEXPORTING, p.pid)
    set_folder_status(task_path, TaskStatus.XEXPORTING, p.pid)
    return p


def export_noquant_model(task_path, save_dir, epoch=None):
    """导出inference模型

    Args:
        task_path(str): 模型训练的路径
        save_dir(str): 导出后的模型保存路径
    """
    output_path = osp.join(task_path, 'output')
    if not osp.exists(osp.join(output_path, 'best_model')):
        raise Exception("未在训练路径{}下发现保存的best_model，导出失败".format(output_path))
    export_status_path = osp.join(task_path, './logs/export')
    safe_clean_folder(export_status_path)
    p = mp.Process(
        target=_call_paddlex_export_infer,
        args=(task_path, save_dir, export_status_path, epoch))
    p.start()
    set_folder_status(export_status_path, TaskStatus.XEXPORTING, p.pid)
    set_folder_status(task_path, TaskStatus.XEXPORTING, p.pid)
    return p


def opt_lite_model(model_path, save_dir=None, place='arm'):
    p = mp.Process(
        target=_call_paddlelite_export_lite,
        args=(model_path, save_dir, place))
    p.start()
    p.join()


def stop_export_task(task_path):
    """停止导出

    Args:
        task_path(str): 从task_path下的'export/XEXPORTING'文件中获取训练的进程id
    Return:
        the export status and message.
    """
    export_status_path = osp.join(task_path, './logs/export')
    if not osp.exists(export_status_path):
        return None, "{}任务目录下没有export文件夹".format(task_path)
    status, message = get_folder_status(export_status_path, True)
    if status == TaskStatus.XEXPORTING:
        pid = int(message)
        is_dead = False
        if not psutil.pid_exists(pid):
            is_dead = True
        else:
            p = psutil.Process(pid)
            if p.status() == 'zombie':
                is_dead = True
        if is_dead:
            status = TaskStatus.XEXPORTFAIL
            message = "导出过程出现异常，请尝试重新评估！"
            set_folder_status(export_status_path, status, message)
        else:
            pkill(pid)
            status = TaskStatus.XEXPORTFAIL
            message = "已停止导出进程！"
            set_folder_status(export_status_path, status, message)
    if status not in [
            TaskStatus.XEXPORTING, TaskStatus.XEXPORTED, TaskStatus.XEXPORTFAIL
    ]:
        raise ValueError("获取到的导出状态异常，{}。".format(status))
    return status, message



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import psutil
import platform


def pkill(pid):
    try:
        parent = psutil.Process(pid)
        for child in parent.children(recursive=True):
            child.kill()
        parent.kill()
    except:
        print("Try to kill process {} failed.".format(pid))


def get_system_info(machine_info={}):
    if machine_info:
        return {'status': 1, 'info': machine_info}
    from .utils import get_gpu_info
    gpu_info, message = get_gpu_info()
    cpu_num = os.environ.get('CPU_NUM', 1)
    sysstr = platform.system()
    machine_info['message'] = message
    machine_info['cpu_num'] = cpu_num
    machine_info['gpu_num'] = gpu_info['gpu_num']
    machine_info['sysstr'] = sysstr
    if gpu_info['gpu_num'] > 0:
        machine_info['driver_version'] = gpu_info['driver_version']
        machine_info['gpu_free_mem'] = gpu_info['mem_free']
    return {'status': 1, 'info': machine_info}


def get_gpu_memory_info(machine_info):
    gpu_mem_infos = list()
    if machine_info['gpu_num'] == 0:
        pass
    else:
        from .utils import get_gpu_info
        gpu_info, message = get_gpu_info()
        for i in range(gpu_info['gpu_num']):
            attr = {
                'free': gpu_info['mem_free'][i],
                'used': gpu_info['mem_used'][i],
                'total': gpu_info['mem_total'][i]
            }
            gpu_mem_infos.append(attr)
    return {'status': 1, 'gpu_mem_infos': gpu_mem_infos}


def get_machine_info(data, machine_info):
    path = None
    if "path" in data:
        path = data['path']
    if path in machine_info:
        return {'status': 1, 'info': machine_info}
    from .utils import get_machine_info
    info = get_machine_info(path)
    machine_info = info
    return {'status': 1, 'info': machine_info}


def get_gpu_memory_size(data):
    """获取显存大小

    Args:
        request(comm.Request): 其中request.params为dict, key包括
        'path' 显卡驱动动态链接库路径
    """
    from .utils import PyNvml
    p = PyNvml()
    p.nvml_init(data['path'])
    count = p.nvml_device_get_count()
    gpu_mem_infos = []
    for i in range(count):
        handler = p.nvml_device_get_handle_by_index(i)
        mem = p.nvml_device_get_memory_info(handler)
        attr = {'free': mem.free, 'used': mem.used, 'total': mem.total}
        gpu_mem_infos.append(attr)
    return {'status': 1, 'gpu_mem_infos': gpu_mem_infos}


def exit_system(monitored_processes):
    while not monitored_processes.empty():
        pid = monitored_processes.get(timeout=0.5)
        print("Try to kill process {}".format(pid))
        pkill(pid)
    return {'status': 1}



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import os.path as osp
import cv2
import numpy as np
import yaml
import multiprocessing as mp
import paddlex
import paddle.fluid as fluid
from paddlex.cv.transforms import build_transforms
from paddlex.cv.models import BaseClassifier
from paddlex.cv.models import PPYOLO, FasterRCNN, MaskRCNN
from paddlex.cv.models import DeepLabv3p
import paddlex.utils.logging as logging


class Predictor:
    def __init__(self,
                 model_dir,
                 use_gpu=True,
                 gpu_id=0,
                 use_mkl=True,
                 mkl_thread_num=4,
                 use_trt=False,
                 use_glog=False,
                 memory_optimize=True,
                 max_trt_batch_size=1):
        """ 创建Paddle Predictor

            Args:
                model_dir: 模型路径（必须是导出的部署或量化模型）
                use_gpu: 是否使用gpu，默认True
                gpu_id: 使用gpu的id，默认0
                use_mkl: 是否使用mkldnn计算库，CPU情况下使用，默认False
                mkl_thread_num: mkldnn计算线程数，默认为4
                use_trt: 是否使用TensorRT，默认False
                use_glog: 是否启用glog日志, 默认False
                memory_optimize: 是否启动内存优化，默认True
                max_trt_batch_size: 在使用TensorRT时配置的最大batch size，默认1
        """
        if not osp.isdir(model_dir):
            raise Exception("[ERROR] Path {} not exist.".format(model_dir))
        if not osp.exists(osp.join(model_dir, "model.yml")):
            raise Exception("There's not model.yml in {}".format(model_dir))
        with open(osp.join(model_dir, "model.yml")) as f:
            self.info = yaml.load(f.read(), Loader=yaml.Loader)

        self.status = self.info['status']

        if self.status != "Quant" and self.status != "Infer":
            raise Exception("[ERROR] Only quantized model or exported "
                            "inference model is supported.")

        self.model_dir = model_dir
        self.model_type = self.info['_Attributes']['model_type']
        self.model_name = self.info['Model']
        self.num_classes = self.info['_Attributes']['num_classes']
        self.labels = self.info['_Attributes']['labels']
        if self.info['Model'] == 'MaskRCNN':
            if self.info['_init_params']['with_fpn']:
                self.mask_head_resolution = 28
            else:
                self.mask_head_resolution = 14
        transforms_mode = self.info.get('TransformsMode', 'RGB')
        if transforms_mode == 'RGB':
            to_rgb = True
        else:
            to_rgb = False
        self.transforms = build_transforms(self.model_type,
                                           self.info['Transforms'], to_rgb)
        self.predictor = self.create_predictor(
            use_gpu, gpu_id, use_mkl, mkl_thread_num, use_trt, use_glog,
            memory_optimize, max_trt_batch_size)
        # 线程池，在模型在预测时用于对输入数据以图片为单位进行并行处理
        # 主要用于batch_predict接口
        thread_num = mp.cpu_count() if mp.cpu_count() < 8 else 8
        self.thread_pool = mp.pool.ThreadPool(thread_num)
        self.input_channel = 3
        if 'input_channel' in self.info['_init_params']:
            self.input_channel = self.info['_init_params']['input_channel']

    def reset_thread_pool(self, thread_num):
        self.thread_pool.close()
        self.thread_pool.join()
        self.thread_pool = mp.pool.ThreadPool(thread_num)

    def create_predictor(self,
                         use_gpu=True,
                         gpu_id=0,
                         use_mkl=False,
                         mkl_thread_num=4,
                         use_trt=False,
                         use_glog=False,
                         memory_optimize=True,
                         max_trt_batch_size=1):
        config = fluid.core.AnalysisConfig(
            os.path.join(self.model_dir, '__model__'),
            os.path.join(self.model_dir, '__params__'))

        if use_gpu:
            # 设置GPU初始显存(单位M)和Device ID
            config.enable_use_gpu(100, gpu_id)
            if use_trt:
                config.enable_tensorrt_engine(
                    workspace_size=1 << 10,
                    max_batch_size=max_trt_batch_size,
                    min_subgraph_size=3,
                    precision_mode=fluid.core.AnalysisConfig.Precision.Float32,
                    use_static=False,
                    use_calib_mode=False)
        else:
            config.disable_gpu()
        if use_mkl and not use_gpu:
            if self.model_name not in ["HRNet", "DeepLabv3p", "PPYOLO"]:
                config.enable_mkldnn()
                config.set_cpu_math_library_num_threads(mkl_thread_num)
            else:
                logging.warning(
                    "HRNet/DeepLabv3p/PPYOLO are not supported for the use of mkldnn\n"
                )
        if use_glog:
            config.enable_glog_info()
        else:
            config.disable_glog_info()
        if memory_optimize:
            config.enable_memory_optim()

        # 开启计算图分析优化，包括OP融合等
        config.switch_ir_optim(True)
        # 关闭feed和fetch OP使用，使用ZeroCopy接口必须设置此项
        config.switch_use_feed_fetch_ops(False)
        predictor = fluid.core.create_paddle_predictor(config)
        return predictor

    def preprocess(self, image, thread_pool=None):
        """ 对图像做预处理

            Args:
                image(list|tuple): 数组中的元素可以是图像路径，也可以是解码后的排列格式为（H，W，C）
                    且类型为float32且为BGR格式的数组。
        """
        res = dict()
        if self.model_type == "classifier":
            im = BaseClassifier._preprocess(
                image,
                self.transforms,
                self.model_type,
                self.model_name,
                thread_pool=thread_pool)
            res['image'] = im
        elif self.model_type == "detector":
            if self.model_name in ["PPYOLO", "YOLOv3"]:
                im, im_size = PPYOLO._preprocess(
                    image,
                    self.transforms,
                    self.model_type,
                    self.model_name,
                    thread_pool=thread_pool,
                    input_channel=self.input_channel)
                res['image'] = im
                res['im_size'] = im_size
            if self.model_name.count('RCNN') > 0:
                im, im_resize_info, im_shape = FasterRCNN._preprocess(
                    image,
                    self.transforms,
                    self.model_type,
                    self.model_name,
                    thread_pool=thread_pool,
                    input_channel=self.input_channel)
                res['image'] = im
                res['im_info'] = im_resize_info
                res['im_shape'] = im_shape
        elif self.model_type == "segmenter":
            im, im_info = DeepLabv3p._preprocess(
                image,
                self.transforms,
                self.model_type,
                self.model_name,
                thread_pool=thread_pool,
                input_channel=self.input_channel)
            res['image'] = im
            res['im_info'] = im_info
        return res

    def postprocess(self,
                    results,
                    topk=1,
                    batch_size=1,
                    im_shape=None,
                    im_info=None):
        """ 对预测结果做后处理

            Args:
                results (list): 预测结果
                topk (int): 分类预测时前k个最大值
                batch_size (int): 预测时图像批量大小
                im_shape (list): MaskRCNN的图像输入大小
                im_info (list)：RCNN系列和分割网络的原图大小
        """

        def offset_to_lengths(lod):
            offset = lod[0]
            lengths = [
                offset[i + 1] - offset[i] for i in range(len(offset) - 1)
            ]
            return [lengths]

        if self.model_type == "classifier":
            true_topk = min(self.num_classes, topk)
            preds = BaseClassifier._postprocess([results[0][0]], true_topk,
                                                self.labels)
        elif self.model_type == "detector":
            res = {'bbox': (results[0][0], offset_to_lengths(results[0][1])), }
            res['im_id'] = (np.array(
                [[i] for i in range(batch_size)]).astype('int32'), [[]])
            if self.model_name in ["PPYOLO", "YOLOv3"]:
                preds = PPYOLO._postprocess(res, batch_size, self.num_classes,
                                            self.labels)
            elif self.model_name == "FasterRCNN":
                preds = FasterRCNN._postprocess(res, batch_size,
                                                self.num_classes, self.labels)
            elif self.model_name == "MaskRCNN":
                res['mask'] = (results[1][0], offset_to_lengths(results[1][1]))
                res['im_shape'] = (im_shape, [])
                preds = MaskRCNN._postprocess(
                    res, batch_size, self.num_classes,
                    self.mask_head_resolution, self.labels)
        elif self.model_type == "segmenter":
            res = [results[0][0], results[1][0]]
            preds = DeepLabv3p._postprocess(res, im_info)
        return preds

    def raw_predict(self, inputs):
        """ 接受预处理过后的数据进行预测

            Args:
                inputs(tuple): 预处理过后的数据
        """
        for k, v in inputs.items():
            try:
                tensor = self.predictor.get_input_tensor(k)
            except:
                continue
            tensor.copy_from_cpu(v)
        self.predictor.zero_copy_run()
        output_names = self.predictor.get_output_names()
        output_results = list()
        for name in output_names:
            output_tensor = self.predictor.get_output_tensor(name)
            output_tensor_lod = output_tensor.lod()
            output_results.append(
                [output_tensor.copy_to_cpu(), output_tensor_lod])
        return output_results

    def predict(self, image, topk=1, transforms=None):
        """ 图片预测

            Args:
                image(str|np.ndarray): 图像路径；或者是解码后的排列格式为（H, W, C）且类型为float32且为BGR格式的数组。
                topk(int): 分类预测时使用，表示预测前topk的结果。
                transforms (paddlex.cls.transforms): 数据预处理操作。
        """
        if transforms is not None:
            self.transforms = transforms
        preprocessed_input = self.preprocess([image])
        model_pred = self.raw_predict(preprocessed_input)
        im_shape = None if 'im_shape' not in preprocessed_input else preprocessed_input[
            'im_shape']
        im_info = None if 'im_info' not in preprocessed_input else preprocessed_input[
            'im_info']
        results = self.postprocess(
            model_pred,
            topk=topk,
            batch_size=1,
            im_shape=im_shape,
            im_info=im_info)

        return results[0]

    def batch_predict(self, image_list, topk=1, transforms=None):
        """ 图片预测

            Args:
                image_list(list|tuple): 对列表（或元组）中的图像同时进行预测，列表中的元素可以是图像路径
                    也可以是解码后的排列格式为（H，W，C）且类型为float32且为BGR格式的数组。

                topk(int): 分类预测时使用，表示预测前topk的结果。
                transforms (paddlex.cls.transforms): 数据预处理操作。
        """
        if transforms is not None:
            self.transforms = transforms
        preprocessed_input = self.preprocess(image_list, self.thread_pool)
        model_pred = self.raw_predict(preprocessed_input)
        im_shape = None if 'im_shape' not in preprocessed_input else preprocessed_input[
            'im_shape']
        im_info = None if 'im_info' not in preprocessed_input else preprocessed_input[
            'im_info']
        results = self.postprocess(
            model_pred,
            topk=topk,
            batch_size=len(image_list),
            im_shape=im_shape,
            im_info=im_info)

        return results



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from .cv.models.slim import prune
from .cv.models.slim import visualize

cal_params_sensitivities = prune.cal_params_sensitivities
visualize = visualize.visualize


def export_quant_model(model,
                       test_dataset,
                       batch_size=1,
                       batch_num=10,
                       save_dir='./quant_model',
                       cache_dir='./temp'):
    model.export_quant_model(
        dataset=test_dataset,
        batch_size=batch_size,
        batch_num=batch_num,
        save_dir=save_dir,
        cache_dir=cache_dir)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import math
import tqdm
import os.path as osp
import numpy as np
from multiprocessing.pool import ThreadPool
import paddle
import paddle.fluid as fluid
from paddle.fluid.layers.learning_rate_scheduler import _decay_step_counter
from paddle.fluid.optimizer import ExponentialMovingAverage
import paddlex.utils.logging as logging
import paddlex
import copy
from paddlex.cv.transforms import arrange_transforms
from paddlex.cv.datasets import generate_minibatch
from .base import BaseAPI
from collections import OrderedDict
from .utils.detection_eval import eval_results, bbox2out


class PPYOLO(BaseAPI):
    """构建PPYOLO，并实现其训练、评估、预测和模型导出。

    Args:
        num_classes (int): 类别数。默认为80。
        backbone (str): PPYOLO的backbone网络，取值范围为['ResNet50_vd_ssld']。默认为'ResNet50_vd_ssld'。
        with_dcn_v2 (bool): Backbone是否使用DCNv2结构。默认为True。
        anchors (list|tuple): anchor框的宽度和高度，为None时表示使用默认值
                    [[10, 13], [16, 30], [33, 23], [30, 61], [62, 45],
                    [59, 119], [116, 90], [156, 198], [373, 326]]。
        anchor_masks (list|tuple): 在计算PPYOLO损失时，使用anchor的mask索引，为None时表示使用默认值
                    [[6, 7, 8], [3, 4, 5], [0, 1, 2]]。
        use_coord_conv (bool): 是否使用CoordConv。默认值为True。
        use_iou_aware (bool): 是否使用IoU Aware分支。默认值为True。
        use_spp (bool): 是否使用Spatial Pyramid Pooling结构。默认值为True。
        use_drop_block (bool): 是否使用Drop Block。默认值为True。
        scale_x_y (float): 调整中心点位置时的系数因子。默认值为1.05。
        use_iou_loss (bool): 是否使用IoU loss。默认值为True。
        use_matrix_nms (bool): 是否使用Matrix NMS。默认值为True。
        ignore_threshold (float): 在计算PPYOLO损失时，IoU大于`ignore_threshold`的预测框的置信度被忽略。默认为0.7。
        nms_score_threshold (float): 检测框的置信度得分阈值，置信度得分低于阈值的框应该被忽略。默认为0.01。
        nms_topk (int): 进行NMS时，根据置信度保留的最大检测框数。默认为1000。
        nms_keep_topk (int): 进行NMS后，每个图像要保留的总检测框数。默认为100。
        nms_iou_threshold (float): 进行NMS时，用于剔除检测框IOU的阈值。默认为0.45。
        label_smooth (bool): 是否使用label smooth。默认值为False。
        train_random_shapes (list|tuple): 训练时从列表中随机选择图像大小。默认值为[320, 352, 384, 416, 448, 480, 512, 544, 576, 608]。
        input_channel (int): 输入图像的通道数量。默认为3。
    """

    def __init__(
            self,
            num_classes=80,
            backbone='ResNet50_vd_ssld',
            with_dcn_v2=True,
            # YOLO Head
            anchors=None,
            anchor_masks=None,
            use_coord_conv=True,
            use_iou_aware=True,
            use_spp=True,
            use_drop_block=True,
            scale_x_y=1.05,
            # PPYOLO Loss
            ignore_threshold=0.7,
            label_smooth=False,
            use_iou_loss=True,
            # NMS
            use_matrix_nms=True,
            nms_score_threshold=0.01,
            nms_topk=1000,
            nms_keep_topk=100,
            nms_iou_threshold=0.45,
            train_random_shapes=[
                320, 352, 384, 416, 448, 480, 512, 544, 576, 608
            ],
            input_channel=3):
        self.init_params = locals()
        super(PPYOLO, self).__init__('detector')
        backbones = ['ResNet50_vd_ssld']
        assert backbone in backbones, "backbone should be one of {}".format(
            backbones)
        self.backbone = backbone
        self.num_classes = num_classes
        self.anchors = anchors
        self.anchor_masks = anchor_masks
        if anchors is None:
            self.anchors = [[10, 13], [16, 30], [33, 23], [30, 61], [62, 45],
                            [59, 119], [116, 90], [156, 198], [373, 326]]
        if anchor_masks is None:
            self.anchor_masks = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]
        self.ignore_threshold = ignore_threshold
        self.nms_score_threshold = nms_score_threshold
        self.nms_topk = nms_topk
        self.nms_keep_topk = nms_keep_topk
        self.nms_iou_threshold = nms_iou_threshold
        self.label_smooth = label_smooth
        self.sync_bn = True
        self.train_random_shapes = train_random_shapes
        self.fixed_input_shape = None
        self.use_fine_grained_loss = False
        if use_coord_conv or use_iou_aware or use_spp or use_drop_block or use_iou_loss:
            self.use_fine_grained_loss = True
        self.use_coord_conv = use_coord_conv
        self.use_iou_aware = use_iou_aware
        self.use_spp = use_spp
        self.use_drop_block = use_drop_block
        self.use_iou_loss = use_iou_loss
        self.scale_x_y = scale_x_y
        self.max_height = 608
        self.max_width = 608
        self.use_matrix_nms = use_matrix_nms
        self.use_ema = False
        self.with_dcn_v2 = with_dcn_v2
        self.input_channel = input_channel

        if paddle.__version__ < '1.8.4' and paddle.__version__ != '0.0.0':
            raise Exception(
                "PPYOLO requires paddlepaddle or paddlepaddle-gpu >= 1.8.4")

    def _get_backbone(self, backbone_name):
        if backbone_name.startswith('ResNet50_vd'):
            backbone = paddlex.cv.nets.ResNet(
                norm_type='sync_bn',
                layers=50,
                freeze_norm=False,
                norm_decay=0.,
                feature_maps=[3, 4, 5],
                freeze_at=0,
                variant='d',
                dcn_v2_stages=[5] if self.with_dcn_v2 else [])
        return backbone

    def build_net(self, mode='train'):
        model = paddlex.cv.nets.detection.YOLOv3(
            backbone=self._get_backbone(self.backbone),
            num_classes=self.num_classes,
            mode=mode,
            anchors=self.anchors,
            anchor_masks=self.anchor_masks,
            ignore_threshold=self.ignore_threshold,
            label_smooth=self.label_smooth,
            nms_score_threshold=self.nms_score_threshold,
            nms_topk=self.nms_topk,
            nms_keep_topk=self.nms_keep_topk,
            nms_iou_threshold=self.nms_iou_threshold,
            fixed_input_shape=self.fixed_input_shape,
            coord_conv=self.use_coord_conv,
            iou_aware=self.use_iou_aware,
            scale_x_y=self.scale_x_y,
            spp=self.use_spp,
            drop_block=self.use_drop_block,
            use_matrix_nms=self.use_matrix_nms,
            use_fine_grained_loss=self.use_fine_grained_loss,
            use_iou_loss=self.use_iou_loss,
            batch_size=getattr(self, 'batch_size_per_gpu', None),
            input_channel=self.input_channel)
        if mode == 'train' and self.use_iou_loss or self.use_iou_aware:
            model.max_height = self.max_height
            model.max_width = self.max_width
        inputs = model.generate_inputs()
        model_out = model.build_net(inputs)
        outputs = OrderedDict([('bbox', model_out)])
        if mode == 'train':
            self.optimizer.minimize(model_out)
            outputs = OrderedDict([('loss', model_out)])
            if self.use_ema:
                global_steps = _decay_step_counter()
                self.ema = ExponentialMovingAverage(
                    self.ema_decay, thres_steps=global_steps)
                self.ema.update()
        return inputs, outputs

    def default_optimizer(self, learning_rate, warmup_steps, warmup_start_lr,
                          lr_decay_epochs, lr_decay_gamma,
                          num_steps_each_epoch):
        if warmup_steps > lr_decay_epochs[0] * num_steps_each_epoch:
            logging.error(
                "In function train(), parameters should satisfy: warmup_steps <= lr_decay_epochs[0]*num_samples_in_train_dataset",
                exit=False)
            logging.error(
                "See this doc for more information: https://github.com/PaddlePaddle/PaddleX/blob/develop/docs/appendix/parameters.md#notice",
                exit=False)
            logging.error(
                "warmup_steps should less than {} or lr_decay_epochs[0] greater than {}, please modify 'lr_decay_epochs' or 'warmup_steps' in train function".
                format(lr_decay_epochs[0] * num_steps_each_epoch, warmup_steps
                       // num_steps_each_epoch))
        boundaries = [b * num_steps_each_epoch for b in lr_decay_epochs]
        values = [(lr_decay_gamma**i) * learning_rate
                  for i in range(len(lr_decay_epochs) + 1)]
        lr_decay = fluid.layers.piecewise_decay(
            boundaries=boundaries, values=values)
        lr_warmup = fluid.layers.linear_lr_warmup(
            learning_rate=lr_decay,
            warmup_steps=warmup_steps,
            start_lr=warmup_start_lr,
            end_lr=learning_rate)
        optimizer = fluid.optimizer.Momentum(
            learning_rate=lr_warmup,
            momentum=0.9,
            regularization=fluid.regularizer.L2DecayRegularizer(5e-04))
        return optimizer

    def train(self,
              num_epochs,
              train_dataset,
              train_batch_size=8,
              eval_dataset=None,
              save_interval_epochs=20,
              log_interval_steps=2,
              save_dir='output',
              pretrain_weights='IMAGENET',
              optimizer=None,
              learning_rate=1.0 / 8000,
              warmup_steps=1000,
              warmup_start_lr=0.0,
              lr_decay_epochs=[213, 240],
              lr_decay_gamma=0.1,
              metric=None,
              use_vdl=False,
              sensitivities_file=None,
              eval_metric_loss=0.05,
              early_stop=False,
              early_stop_patience=5,
              resume_checkpoint=None,
              use_ema=True,
              ema_decay=0.9998):
        """训练。

        Args:
            num_epochs (int): 训练迭代轮数。
            train_dataset (paddlex.datasets): 训练数据读取器。
            train_batch_size (int): 训练数据batch大小。目前检测仅支持单卡评估，训练数据batch大小与显卡
                数量之商为验证数据batch大小。默认值为8。
            eval_dataset (paddlex.datasets): 验证数据读取器。
            save_interval_epochs (int): 模型保存间隔（单位：迭代轮数）。默认为20。
            log_interval_steps (int): 训练日志输出间隔（单位：迭代次数）。默认为10。
            save_dir (str): 模型保存路径。默认值为'output'。
            pretrain_weights (str): 若指定为路径时，则加载路径下预训练模型；若为字符串'IMAGENET'，
                则自动下载在ImageNet图片数据上预训练的模型权重；若为字符串'COCO'，
                则自动下载在COCO数据集上预训练的模型权重；若为None，则不使用预训练模型。默认为'IMAGENET'。
            optimizer (paddle.fluid.optimizer): 优化器。当该参数为None时，使用默认优化器：
                fluid.layers.piecewise_decay衰减策略，fluid.optimizer.Momentum优化方法。
            learning_rate (float): 默认优化器的学习率。默认为1.0/8000。
            warmup_steps (int):  默认优化器进行warmup过程的步数。默认为1000。
            warmup_start_lr (int): 默认优化器warmup的起始学习率。默认为0.0。
            lr_decay_epochs (list): 默认优化器的学习率衰减轮数。默认为[213, 240]。
            lr_decay_gamma (float): 默认优化器的学习率衰减率。默认为0.1。
            metric (bool): 训练过程中评估的方式，取值范围为['COCO', 'VOC']。默认值为None。
            use_vdl (bool): 是否使用VisualDL进行可视化。默认值为False。
            sensitivities_file (str): 若指定为路径时，则加载路径下敏感度信息进行裁剪；若为字符串'DEFAULT'，
                则自动下载在ImageNet图片数据上获得的敏感度信息进行裁剪；若为None，则不进行裁剪。默认为None。
            eval_metric_loss (float): 可容忍的精度损失。默认为0.05。
            early_stop (bool): 是否使用提前终止训练策略。默认值为False。
            early_stop_patience (int): 当使用提前终止训练策略时，如果验证集精度在`early_stop_patience`个epoch内
                连续下降或持平，则终止训练。默认值为5。
            resume_checkpoint (str): 恢复训练时指定上次训练保存的模型路径。若为None，则不会恢复训练。默认值为None。
            use_ema (bool): 是否使用指数衰减计算参数的滑动平均值。默认值为True。
            ema_decay (float): 指数衰减率。默认值为0.9998。

        Raises:
            ValueError: 评估类型不在指定列表中。
            ValueError: 模型从inference model进行加载。
        """
        if not self.trainable:
            raise ValueError("Model is not trainable from load_model method.")
        if metric is None:
            if isinstance(train_dataset, paddlex.datasets.CocoDetection):
                metric = 'COCO'
            elif isinstance(train_dataset, paddlex.datasets.VOCDetection) or \
                    isinstance(train_dataset, paddlex.datasets.EasyDataDet):
                metric = 'VOC'
            else:
                raise ValueError(
                    "train_dataset should be datasets.VOCDetection or datasets.COCODetection or datasets.EasyDataDet."
                )
        assert metric in ['COCO', 'VOC'], "Metric only support 'VOC' or 'COCO'"
        self.metric = metric

        self.labels = train_dataset.labels
        # 构建训练网络
        if optimizer is None:
            # 构建默认的优化策略
            num_steps_each_epoch = train_dataset.num_samples // train_batch_size
            optimizer = self.default_optimizer(
                learning_rate=learning_rate,
                warmup_steps=warmup_steps,
                warmup_start_lr=warmup_start_lr,
                lr_decay_epochs=lr_decay_epochs,
                lr_decay_gamma=lr_decay_gamma,
                num_steps_each_epoch=num_steps_each_epoch)
        self.optimizer = optimizer
        self.use_ema = use_ema
        self.ema_decay = ema_decay

        self.batch_size_per_gpu = self._get_single_card_bs(train_batch_size)
        if self.use_fine_grained_loss:
            for transform in train_dataset.transforms.transforms:
                if isinstance(transform, paddlex.det.transforms.Resize):
                    self.max_height = transform.target_size
                    self.max_width = transform.target_size
                    break
        if train_dataset.transforms.batch_transforms is None:
            train_dataset.transforms.batch_transforms = list()
        define_random_shape = False
        for bt in train_dataset.transforms.batch_transforms:
            if isinstance(bt, paddlex.det.transforms.BatchRandomShape):
                define_random_shape = True
        if not define_random_shape:
            if isinstance(self.train_random_shapes,
                          (list, tuple)) and len(self.train_random_shapes) > 0:
                train_dataset.transforms.batch_transforms.append(
                    paddlex.det.transforms.BatchRandomShape(
                        random_shapes=self.train_random_shapes))
                if self.use_fine_grained_loss:
                    self.max_height = max(self.max_height,
                                          max(self.train_random_shapes))
                    self.max_width = max(self.max_width,
                                         max(self.train_random_shapes))
        if self.use_fine_grained_loss:
            define_generate_target = False
            for bt in train_dataset.transforms.batch_transforms:
                if isinstance(bt, paddlex.det.transforms.GenerateYoloTarget):
                    define_generate_target = True
            if not define_generate_target:
                train_dataset.transforms.batch_transforms.append(
                    paddlex.det.transforms.GenerateYoloTarget(
                        anchors=self.anchors,
                        anchor_masks=self.anchor_masks,
                        num_classes=self.num_classes,
                        downsample_ratios=[32, 16, 8]))
        # 构建训练、验证、预测网络
        self.build_program()
        # 初始化网络权重
        self.net_initialize(
            startup_prog=fluid.default_startup_program(),
            pretrain_weights=pretrain_weights,
            save_dir=save_dir,
            sensitivities_file=sensitivities_file,
            eval_metric_loss=eval_metric_loss,
            resume_checkpoint=resume_checkpoint)
        # 训练
        self.train_loop(
            num_epochs=num_epochs,
            train_dataset=train_dataset,
            train_batch_size=train_batch_size,
            eval_dataset=eval_dataset,
            save_interval_epochs=save_interval_epochs,
            log_interval_steps=log_interval_steps,
            save_dir=save_dir,
            use_vdl=use_vdl,
            early_stop=early_stop,
            early_stop_patience=early_stop_patience)

    def evaluate(self,
                 eval_dataset,
                 batch_size=1,
                 epoch_id=None,
                 metric=None,
                 return_details=False):
        """评估。

        Args:
            eval_dataset (paddlex.datasets): 验证数据读取器。
            batch_size (int): 验证数据批大小。默认为1。
            epoch_id (int): 当前评估模型所在的训练轮数。
            metric (bool): 训练过程中评估的方式，取值范围为['COCO', 'VOC']。默认为None，
                根据用户传入的Dataset自动选择，如为VOCDetection，则metric为'VOC';
                如为COCODetection，则metric为'COCO'。
            return_details (bool): 是否返回详细信息。

        Returns:
            tuple (metrics, eval_details) | dict (metrics): 当return_details为True时，返回(metrics, eval_details)，
                当return_details为False时，返回metrics。metrics为dict，包含关键字：'bbox_mmap'或者’bbox_map‘，
                分别表示平均准确率平均值在各个IoU阈值下的结果取平均值的结果（mmAP）、平均准确率平均值（mAP）。
                eval_details为dict，包含bbox和gt两个关键字。其中关键字bbox的键值是一个列表，列表中每个元素代表一个预测结果，
                一个预测结果是一个由图像id，预测框类别id, 预测框坐标，预测框得分组成的列表。而关键字gt的键值是真实标注框的相关信息。
        """
        input_channel = getattr(self, 'input_channel', 3)
        arrange_transforms(
            model_type=self.model_type,
            class_name=self.__class__.__name__,
            transforms=eval_dataset.transforms,
            mode='eval',
            input_channel=input_channel)
        if metric is None:
            if hasattr(self, 'metric') and self.metric is not None:
                metric = self.metric
            else:
                if isinstance(eval_dataset, paddlex.datasets.CocoDetection):
                    metric = 'COCO'
                elif isinstance(eval_dataset, paddlex.datasets.VOCDetection):
                    metric = 'VOC'
                else:
                    raise Exception(
                        "eval_dataset should be datasets.VOCDetection or datasets.COCODetection."
                    )
        assert metric in ['COCO', 'VOC'], "Metric only support 'VOC' or 'COCO'"

        total_steps = math.ceil(eval_dataset.num_samples * 1.0 / batch_size)
        results = list()

        data_generator = eval_dataset.generator(
            batch_size=batch_size, drop_last=False)
        logging.info(
            "Start to evaluating(total_samples={}, total_steps={})...".format(
                eval_dataset.num_samples, total_steps))
        for step, data in tqdm.tqdm(
                enumerate(data_generator()), total=total_steps):
            images = np.array([d[0] for d in data])
            im_sizes = np.array([d[1] for d in data])
            feed_data = {'image': images, 'im_size': im_sizes}
            with fluid.scope_guard(self.scope):
                outputs = self.exe.run(
                    self.test_prog,
                    feed=[feed_data],
                    fetch_list=list(self.test_outputs.values()),
                    return_numpy=False)
            res = {
                'bbox': (np.array(outputs[0]),
                         outputs[0].recursive_sequence_lengths())
            }
            res_id = [np.array([d[2]]) for d in data]
            res['im_id'] = (res_id, [])
            if metric == 'VOC':
                res_gt_box = [d[3].reshape(-1, 4) for d in data]
                res_gt_label = [d[4].reshape(-1, 1) for d in data]
                res_is_difficult = [d[5].reshape(-1, 1) for d in data]
                res_id = [np.array([d[2]]) for d in data]
                res['gt_box'] = (res_gt_box, [])
                res['gt_label'] = (res_gt_label, [])
                res['is_difficult'] = (res_is_difficult, [])
            results.append(res)
            logging.debug("[EVAL] Epoch={}, Step={}/{}".format(epoch_id, step +
                                                               1, total_steps))
        box_ap_stats, eval_details = eval_results(
            results, metric, eval_dataset.coco_gt, with_background=False)
        evaluate_metrics = OrderedDict(
            zip(['bbox_mmap'
                 if metric == 'COCO' else 'bbox_map'], box_ap_stats))
        if return_details:
            return evaluate_metrics, eval_details
        return evaluate_metrics

    @staticmethod
    def _preprocess(images,
                    transforms,
                    model_type,
                    class_name,
                    thread_pool=None,
                    input_channel=3):
        arrange_transforms(
            model_type=model_type,
            class_name=class_name,
            transforms=transforms,
            mode='test',
            input_channel=input_channel)
        if thread_pool is not None:
            batch_data = thread_pool.map(transforms, images)
        else:
            batch_data = list()
            for image in images:
                batch_data.append(transforms(image))
        padding_batch = generate_minibatch(batch_data)
        im = np.array(
            [data[0] for data in padding_batch],
            dtype=padding_batch[0][0].dtype)
        im_size = np.array([data[1] for data in padding_batch], dtype=np.int32)

        return im, im_size

    @staticmethod
    def _postprocess(res, batch_size, num_classes, labels):
        clsid2catid = dict({i: i for i in range(num_classes)})
        xywh_results = bbox2out([res], clsid2catid)
        preds = [[] for i in range(batch_size)]
        for xywh_res in xywh_results:
            image_id = xywh_res['image_id']
            del xywh_res['image_id']
            xywh_res['category'] = labels[xywh_res['category_id']]
            preds[image_id].append(xywh_res)

        return preds

    def predict(self, img_file, transforms=None):
        """预测。

        Args:
            img_file (str|np.ndarray): 预测图像路径，或者是解码后的排列格式为（H, W, C）且类型为float32且为BGR格式的数组。
            transforms (paddlex.det.transforms): 数据预处理操作。

        Returns:
            list: 预测结果列表，每个预测结果由预测框类别标签、
              预测框类别名称、预测框坐标(坐标格式为[xmin, ymin, w, h]）、
              预测框得分组成。
        """
        if transforms is None and not hasattr(self, 'test_transforms'):
            raise Exception("transforms need to be defined, now is None.")
        if isinstance(img_file, (str, np.ndarray)):
            images = [img_file]
        else:
            raise Exception("img_file must be str/np.ndarray")

        if transforms is None:
            transforms = self.test_transforms
        input_channel = getattr(self, 'input_channel', 3)
        im, im_size = PPYOLO._preprocess(
            images,
            transforms,
            self.model_type,
            self.__class__.__name__,
            input_channel=input_channel)

        with fluid.scope_guard(self.scope):
            result = self.exe.run(self.test_prog,
                                  feed={'image': im,
                                        'im_size': im_size},
                                  fetch_list=list(self.test_outputs.values()),
                                  return_numpy=False,
                                  use_program_cache=True)

        res = {
            k: (np.array(v), v.recursive_sequence_lengths())
            for k, v in zip(list(self.test_outputs.keys()), result)
        }
        res['im_id'] = (np.array(
            [[i] for i in range(len(images))]).astype('int32'), [[]])
        preds = PPYOLO._postprocess(res,
                                    len(images), self.num_classes, self.labels)
        return preds[0]

    def batch_predict(self, img_file_list, transforms=None):
        """预测。

        Args:
            img_file_list (list|tuple): 对列表（或元组）中的图像同时进行预测，列表中的元素可以是图像路径，也可以是解码后的排列格式为（H，W，C）
                且类型为float32且为BGR格式的数组。
            transforms (paddlex.det.transforms): 数据预处理操作。
        Returns:
            list: 每个元素都为列表，表示各图像的预测结果。在各图像的预测结果列表中，每个预测结果由预测框类别标签、
              预测框类别名称、预测框坐标(坐标格式为[xmin, ymin, w, h]）、
              预测框得分组成。
        """
        if transforms is None and not hasattr(self, 'test_transforms'):
            raise Exception("transforms need to be defined, now is None.")

        if not isinstance(img_file_list, (list, tuple)):
            raise Exception("im_file must be list/tuple")

        if transforms is None:
            transforms = self.test_transforms
        input_channel = getattr(self, 'input_channel', 3)
        im, im_size = PPYOLO._preprocess(
            img_file_list,
            transforms,
            self.model_type,
            self.__class__.__name__,
            self.thread_pool,
            input_channel=input_channel)
        im, im_size = PPYOLO._preprocess(
            img_file_list, transforms, self.model_type,
            self.__class__.__name__, self.thread_pool)

        with fluid.scope_guard(self.scope):
            result = self.exe.run(self.test_prog,
                                  feed={'image': im,
                                        'im_size': im_size},
                                  fetch_list=list(self.test_outputs.values()),
                                  return_numpy=False,
                                  use_program_cache=True)

        res = {
            k: (np.array(v), v.recursive_sequence_lengths())
            for k, v in zip(list(self.test_outputs.keys()), result)
        }
        res['im_id'] = (np.array(
            [[i] for i in range(len(img_file_list))]).astype('int32'), [[]])
        preds = PPYOLO._postprocess(res,
                                    len(img_file_list), self.num_classes,
                                    self.labels)
        return preds



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import numpy as np
from scipy.sparse import csr_matrix


class ConfusionMatrix(object):
    """
        Confusion Matrix for segmentation evaluation
    """

    def __init__(self, num_classes=2, streaming=False):
        self.confusion_matrix = np.zeros(
            [num_classes, num_classes], dtype='int64')
        self.num_classes = num_classes
        self.streaming = streaming

    def calculate(self, pred, label, ignore=None):
        # If not in streaming mode, clear matrix everytime when call `calculate`
        if not self.streaming:
            self.zero_matrix()

        label = np.transpose(label, (0, 2, 3, 1))
        ignore = np.transpose(ignore, (0, 2, 3, 1))
        mask = np.array(ignore) == 1

        label = np.asarray(label)[mask]
        pred = np.asarray(pred)[mask]
        one = np.ones_like(pred)
        # Accumuate ([row=label, col=pred], 1) into sparse matrix
        spm = csr_matrix(
            (one, (label, pred)), shape=(self.num_classes, self.num_classes))
        spm = spm.todense()
        self.confusion_matrix += spm

    def zero_matrix(self):
        """ Clear confusion matrix """
        self.confusion_matrix = np.zeros(
            [self.num_classes, self.num_classes], dtype='int64')

    def mean_iou(self):
        iou_list = []
        avg_iou = 0
        # TODO: use numpy sum axis api to simpliy
        vji = np.zeros(self.num_classes, dtype=int)
        vij = np.zeros(self.num_classes, dtype=int)
        for j in range(self.num_classes):
            v_j = 0
            for i in range(self.num_classes):
                v_j += self.confusion_matrix[j][i]
            vji[j] = v_j

        for i in range(self.num_classes):
            v_i = 0
            for j in range(self.num_classes):
                v_i += self.confusion_matrix[j][i]
            vij[i] = v_i

        for c in range(self.num_classes):
            total = vji[c] + vij[c] - self.confusion_matrix[c][c]
            if total == 0:
                iou = 0
            else:
                iou = float(self.confusion_matrix[c][c]) / total
            avg_iou += iou
            iou_list.append(iou)
        avg_iou = float(avg_iou) / float(self.num_classes)
        return np.array(iou_list), avg_iou

    def accuracy(self):
        total = self.confusion_matrix.sum()
        total_right = 0
        for c in range(self.num_classes):
            total_right += self.confusion_matrix[c][c]
        if total == 0:
            avg_acc = 0
        else:
            avg_acc = float(total_right) / total

        vij = np.zeros(self.num_classes, dtype=int)
        for i in range(self.num_classes):
            v_i = 0
            for j in range(self.num_classes):
                v_i += self.confusion_matrix[j][i]
            vij[i] = v_i

        acc_list = []
        for c in range(self.num_classes):
            if vij[c] == 0:
                acc = 0
            else:
                acc = self.confusion_matrix[c][c] / float(vij[c])
            acc_list.append(acc)
        return np.array(acc_list), avg_acc

    def kappa(self):
        vji = np.zeros(self.num_classes)
        vij = np.zeros(self.num_classes)
        for j in range(self.num_classes):
            v_j = 0
            for i in range(self.num_classes):
                v_j += self.confusion_matrix[j][i]
            vji[j] = v_j

        for i in range(self.num_classes):
            v_i = 0
            for j in range(self.num_classes):
                v_i += self.confusion_matrix[j][i]
            vij[i] = v_i

        total = self.confusion_matrix.sum()

        # avoid spillovers
        # TODO: is it reasonable to hard code 10000.0?
        total = float(total) / 10000.0
        vji = vji / 10000.0
        vij = vij / 10000.0

        tp = 0
        tc = 0
        for c in range(self.num_classes):
            tp += vji[c] * vij[c]
            tc += self.confusion_matrix[c][c]

        tc = tc / 10000.0
        pe = tp / (total * total)
        po = tc / total

        kappa = (po - pe) / (1 - pe)
        return kappa

    def f1_score(self):
        f1score_list = []
        # TODO: use numpy sum axis api to simpliy
        vji = np.zeros(self.num_classes, dtype=int)
        vij = np.zeros(self.num_classes, dtype=int)
        for j in range(self.num_classes):
            v_j = 0
            for i in range(self.num_classes):
                v_j += self.confusion_matrix[j][i]
            vji[j] = v_j

        for i in range(self.num_classes):
            v_i = 0
            for j in range(self.num_classes):
                v_i += self.confusion_matrix[j][i]
            vij[i] = v_i

        for c in range(self.num_classes):
            if vji[c] == 0:
                precision = 0
            else:
                precision = self.confusion_matrix[c][c] / vji[c]
            if vij[c] == 0:
                recall = 0
            else:
                recall = self.confusion_matrix[c][c] / vij[c]
            if recall + precision <= 1e-06:
                f1score = 0
            else:
                f1score = 2 * precision * recall / (recall + precision)
            f1score_list.append(f1score)
        return np.array(f1score_list)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path as osp
import numpy as np
import glob
import tqdm

from paddlex.interpret.as_data_reader.readers import read_image
import paddlex.utils.logging as logging
from . import lime_base
from ._session_preparation import compute_features_for_kmeans, gen_user_home
import paddlex.utils.logging as logging


def load_kmeans_model(fname):
    import pickle
    with open(fname, 'rb') as f:
        kmeans_model = pickle.load(f)

    return kmeans_model


def combine_normlime_and_lime(lime_weights, g_weights):
    pred_labels = lime_weights.keys()
    combined_weights = {y: [] for y in pred_labels}

    for y in pred_labels:
        normlized_lime_weights_y = lime_weights[y]
        lime_weights_dict = {
            tuple_w[0]: tuple_w[1]
            for tuple_w in normlized_lime_weights_y
        }

        normlized_g_weight_y = g_weights[y]
        normlime_weights_dict = {
            tuple_w[0]: tuple_w[1]
            for tuple_w in normlized_g_weight_y
        }

        combined_weights[y] = [
            (seg_k, lime_weights_dict[seg_k] * normlime_weights_dict[seg_k])
            for seg_k in lime_weights_dict.keys()
        ]

        combined_weights[y] = sorted(
            combined_weights[y], key=lambda x: np.abs(x[1]), reverse=True)

    return combined_weights


def avg_using_superpixels(features, segments):
    one_list = np.zeros((len(np.unique(segments)), features.shape[2]))
    for x in np.unique(segments):
        one_list[x] = np.mean(features[segments == x], axis=0)

    return one_list


def centroid_using_superpixels(features, segments):
    from skimage.measure import regionprops
    regions = regionprops(segments + 1)
    one_list = np.zeros((len(np.unique(segments)), features.shape[2]))
    for i, r in enumerate(regions):
        one_list[i] = features[int(r.centroid[0] + 0.5), int(r.centroid[1] +
                                                             0.5), :]
    return one_list


def get_feature_for_kmeans(feature_map, segments):
    from sklearn.preprocessing import normalize
    centroid_feature = centroid_using_superpixels(feature_map, segments)
    avg_feature = avg_using_superpixels(feature_map, segments)
    x = np.concatenate((centroid_feature, avg_feature), axis=-1)
    x = normalize(x)
    return x


def precompute_normlime_weights(list_data_,
                                predict_fn,
                                num_samples=3000,
                                batch_size=50,
                                save_dir='./tmp'):
    # save lime weights and kmeans cluster labels
    precompute_lime_weights(list_data_, predict_fn, num_samples, batch_size,
                            save_dir)

    # load precomputed results, compute normlime weights and save.
    fname_list = glob.glob(
        os.path.join(save_dir, 'lime_weights_s{}*.npy'.format(num_samples)))
    return compute_normlime_weights(fname_list, save_dir, num_samples)


def save_one_lime_predict_and_kmean_labels(lime_all_weights, image_pred_labels,
                                           cluster_labels, save_path):

    lime_weights = {}
    for label in image_pred_labels:
        lime_weights[label] = lime_all_weights[label]

    for_normlime_weights = {
        'lime_weights':
        lime_weights,  # a dict: class_label: (seg_label, weight)
        'cluster': cluster_labels  # a list with segments as indices.
    }

    np.save(save_path, for_normlime_weights)


def precompute_lime_weights(list_data_, predict_fn, num_samples, batch_size,
                            save_dir):
    root_path = gen_user_home()
    root_path = osp.join(root_path, '.paddlex')
    h_pre_models = osp.join(root_path, "pre_models")
    if not osp.exists(h_pre_models):
        if not osp.exists(root_path):
            os.makedirs(root_path)
        url = "https://bj.bcebos.com/paddlex/interpret/pre_models.tar.gz"
        pdx.utils.download_and_decompress(url, path=root_path)
    h_pre_models_kmeans = osp.join(h_pre_models, "kmeans_model.pkl")
    kmeans_model = load_kmeans_model(h_pre_models_kmeans)

    for data_index, each_data_ in enumerate(list_data_):
        if isinstance(each_data_, str):
            save_path = "lime_weights_s{}_{}.npy".format(
                num_samples, each_data_.split('/')[-1].split('.')[0])
            save_path = os.path.join(save_dir, save_path)
        else:
            save_path = "lime_weights_s{}_{}.npy".format(num_samples,
                                                         data_index)
            save_path = os.path.join(save_dir, save_path)

        if os.path.exists(save_path):
            logging.info(
                save_path + ' exists, not computing this one.', use_color=True)
            continue
        img_file_name = each_data_ if isinstance(each_data_,
                                                 str) else data_index
        logging.info(
            'processing ' + img_file_name + ' [{}/{}]'.format(data_index,
                                                              len(list_data_)),
            use_color=True)

        image_show = read_image(each_data_)
        result = predict_fn(image_show)
        result = result[0]  # only one image here.

        if abs(np.sum(result) - 1.0) > 1e-4:
            # softmax
            exp_result = np.exp(result)
            probability = exp_result / np.sum(exp_result)
        else:
            probability = result

        pred_label = np.argsort(probability)[::-1]

        # top_k = argmin(top_n) > threshold
        threshold = 0.05
        top_k = 0
        for l in pred_label:
            if probability[l] < threshold or top_k == 5:
                break
            top_k += 1

        if top_k == 0:
            top_k = 1

        pred_label = pred_label[:top_k]

        algo = lime_base.LimeImageInterpreter()
        interpreter = algo.interpret_instance(
            image_show[0],
            predict_fn,
            pred_label,
            0,
            num_samples=num_samples,
            batch_size=batch_size)

        X = get_feature_for_kmeans(
            compute_features_for_kmeans(image_show).transpose((1, 2, 0)),
            interpreter.segments)
        try:
            cluster_labels = kmeans_model.predict(X)
        except AttributeError:
            from sklearn.metrics import pairwise_distances_argmin_min
            cluster_labels, _ = pairwise_distances_argmin_min(
                X, kmeans_model.cluster_centers_)
        save_one_lime_predict_and_kmean_labels(
            interpreter.local_weights, pred_label, cluster_labels, save_path)


def compute_normlime_weights(a_list_lime_fnames, save_dir, lime_num_samples):
    normlime_weights_all_labels = {}

    for f in a_list_lime_fnames:
        try:
            lime_weights_and_cluster = np.load(f, allow_pickle=True).item()
            lime_weights = lime_weights_and_cluster['lime_weights']
            cluster = lime_weights_and_cluster['cluster']
        except:
            logging.info('When loading precomputed LIME result, skipping' +
                         str(f))
            continue
        logging.info('Loading precomputed LIME result,' + str(f))
        pred_labels = lime_weights.keys()
        for y in pred_labels:
            normlime_weights = normlime_weights_all_labels.get(y, {})
            w_f_y = [abs(w[1]) for w in lime_weights[y]]
            w_f_y_l1norm = sum(w_f_y)

            for w in lime_weights[y]:
                seg_label = w[0]
                weight = w[1] * w[1] / w_f_y_l1norm
                a = normlime_weights.get(cluster[seg_label], [])
                a.append(weight)
                normlime_weights[cluster[seg_label]] = a

            normlime_weights_all_labels[y] = normlime_weights

    # compute normlime
    for y in normlime_weights_all_labels:
        normlime_weights = normlime_weights_all_labels.get(y, {})
        for k in normlime_weights:
            normlime_weights[k] = sum(normlime_weights[k]) / len(
                normlime_weights[k])

    # check normlime
    if len(normlime_weights_all_labels.keys()) < max(
            normlime_weights_all_labels.keys()) + 1:
        logging.info(
            "\n" + \
            "Warning: !!! \n" + \
            "There are at least {} classes, ".format(max(normlime_weights_all_labels.keys()) + 1) + \
            "but the NormLIME has results of only {} classes. \n".format(len(normlime_weights_all_labels.keys())) + \
            "It may have cause unstable results in the later computation" + \
            " but can be improved by computing more test samples." + \
            "\n"
        )

    n = 0
    f_out = 'normlime_weights_s{}_samples_{}-{}.npy'.format(
        lime_num_samples, len(a_list_lime_fnames), n)
    while os.path.exists(os.path.join(save_dir, f_out)):
        n += 1
        f_out = 'normlime_weights_s{}_samples_{}-{}.npy'.format(
            lime_num_samples, len(a_list_lime_fnames), n)
        continue

    np.save(os.path.join(save_dir, f_out), normlime_weights_all_labels)
    return os.path.join(save_dir, f_out)


def precompute_global_classifier(dataset,
                                 predict_fn,
                                 save_path,
                                 batch_size=50,
                                 max_num_samples=1000):
    from sklearn.linear_model import LogisticRegression

    root_path = gen_user_home()
    root_path = osp.join(root_path, '.paddlex')
    h_pre_models = osp.join(root_path, "pre_models")
    if not osp.exists(h_pre_models):
        if not osp.exists(root_path):
            os.makedirs(root_path)
        url = "https://bj.bcebos.com/paddlex/interpret/pre_models.tar.gz"
        pdx.utils.download_and_decompress(url, path=root_path)
    h_pre_models_kmeans = osp.join(h_pre_models, "kmeans_model.pkl")
    kmeans_model = load_kmeans_model(h_pre_models_kmeans)

    image_list = []
    for item in dataset.file_list:
        image_list.append(item[0])

    x_data = []
    y_labels = []

    num_features = len(kmeans_model.cluster_centers_)

    logging.info(
        "Initialization for NormLIME: Computing each sample in the test list.",
        use_color=True)

    for each_data_ in tqdm.tqdm(image_list):
        x_data_i = np.zeros((num_features))
        image_show = read_image(each_data_)
        result = predict_fn(image_show)
        result = result[0]  # only one image here.
        c = compute_features_for_kmeans(image_show).transpose((1, 2, 0))

        segments = np.zeros((image_show.shape[1], image_show.shape[2]),
                            np.int32)
        num_blocks = 10
        height_per_i = segments.shape[0] // num_blocks + 1
        width_per_i = segments.shape[1] // num_blocks + 1

        for i in range(segments.shape[0]):
            for j in range(segments.shape[1]):
                segments[i,
                         j] = i // height_per_i * num_blocks + j // width_per_i

        # segments = quickshift(image_show[0], sigma=1)
        X = get_feature_for_kmeans(c, segments)

        try:
            cluster_labels = kmeans_model.predict(X)
        except AttributeError:
            from sklearn.metrics import pairwise_distances_argmin_min
            cluster_labels, _ = pairwise_distances_argmin_min(
                X, kmeans_model.cluster_centers_)

        for c in cluster_labels:
            x_data_i[c] = 1

        # x_data_i /= len(cluster_labels)

        pred_y_i = np.argmax(result)
        y_labels.append(pred_y_i)
        x_data.append(x_data_i)

    if len(np.unique(y_labels)) < 2:
        logging.info("Warning: The test samples in the dataset is limited.\n \
                     NormLIME may have no effect on the results.\n \
                     Try to add more test samples, or see the results of LIME.")
        num_classes = np.max(np.unique(y_labels)) + 1
        normlime_weights_all_labels = {}
        for class_index in range(num_classes):
            w = np.ones((num_features)) / num_features
            normlime_weights_all_labels[class_index] = {
                i: wi
                for i, wi in enumerate(w)
            }
        logging.info("Saving the computed normlime_weights in {}".format(
            save_path))

        np.save(save_path, normlime_weights_all_labels)
        return save_path

    clf = LogisticRegression(multi_class='multinomial', max_iter=1000)
    clf.fit(x_data, y_labels)

    num_classes = np.max(np.unique(y_labels)) + 1
    normlime_weights_all_labels = {}

    if len(y_labels) / len(np.unique(y_labels)) < 3:
        logging.info("Warning: The test samples in the dataset is limited.\n \
                     NormLIME may have no effect on the results.\n \
                     Try to add more test samples, or see the results of LIME.")

    if len(np.unique(y_labels)) == 2:
        # binary: clf.coef_ has shape of [1, num_features]
        for class_index in range(num_classes):
            if class_index not in clf.classes_:
                w = np.ones((num_features)) / num_features
                normlime_weights_all_labels[class_index] = {
                    i: wi
                    for i, wi in enumerate(w)
                }
                continue

            if clf.classes_[0] == class_index:
                w = -clf.coef_[0]
            else:
                w = clf.coef_[0]

            # softmax
            w = w - np.max(w)
            exp_w = np.exp(w * 10)
            w = exp_w / np.sum(exp_w)

            normlime_weights_all_labels[class_index] = {
                i: wi
                for i, wi in enumerate(w)
            }
    else:
        # clf.coef_ has shape of [len(np.unique(y_labels)), num_features]
        for class_index in range(num_classes):
            if class_index not in clf.classes_:
                w = np.ones((num_features)) / num_features
                normlime_weights_all_labels[class_index] = {
                    i: wi
                    for i, wi in enumerate(w)
                }
                continue

            coef_class_index = np.where(clf.classes_ == class_index)[0][0]
            w = clf.coef_[coef_class_index]

            # softmax
            w = w - np.max(w)
            exp_w = np.exp(w * 10)
            w = exp_w / np.sum(exp_w)

            normlime_weights_all_labels[class_index] = {
                i: wi
                for i, wi in enumerate(w)
            }

    logging.info("Saving the computed normlime_weights in {}".format(
        save_path))
    np.save(save_path, normlime_weights_all_labels)

    return save_path



# copytrue (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import cv2
import math
import xml.etree.ElementTree as ET
from PIL import Image


def resize_img(img):
    """ 调整图片尺寸

    Args:
        img: 图片信息
    """
    h, w = img.shape[:2]
    min_size = 580

    if w >= h and w > min_size:
        new_w = min_size
        new_h = new_w * h / w
    elif h >= w and h > min_size:
        new_h = min_size
        new_w = new_h * w / h
    else:
        new_h = h
        new_w = w
    new_img = cv2.resize(
        img, (int(new_w), int(new_h)), interpolation=cv2.INTER_CUBIC)

    scale_value = new_w / w
    return new_img, scale_value


def plot_det_label(image, anno, labels):
    """ 目标检测类型生成标注图

    Args:
        image: 图片路径
        anno: 图片标注
        labels: 图片所属数据集的类别信息
    """
    catid2color = {}
    img = cv2.imread(image)
    img, scale_value = resize_img(img)
    tree = ET.parse(anno)
    objs = tree.findall('object')
    color_map = get_color_map_list(len(labels) + 1)
    for i, obj in enumerate(objs):
        cname = obj.find('name').text
        catid = labels.index(cname)
        if cname not in labels:
            continue
        xmin = int(float(obj.find('bndbox').find('xmin').text) * scale_value)
        ymin = int(float(obj.find('bndbox').find('ymin').text) * scale_value)
        xmax = int(float(obj.find('bndbox').find('xmax').text) * scale_value)
        ymax = int(float(obj.find('bndbox').find('ymax').text) * scale_value)

        if catid not in catid2color:
            catid2color[catid] = color_map[catid + 1]
        color = tuple(catid2color[catid])
        img = draw_rectangle_and_cname(img, xmin, ymin, xmax, ymax, cname,
                                       color)
    return img


def plot_seg_label(anno):
    """ 语义分割类型生成标注图

    Args:
        anno: 图片标注
    """
    label = pil_imread(anno)
    pse_label = gray2pseudo(label)
    return pse_label


def plot_insseg_label(image, anno, labels, alpha=0.7):
    """ 实例分割类型生成标注图

    Args:
        image: 图片路径
        anno: 图片标注
        labels: 图片所属数据集的类别信息
    """
    anno = np.load(anno, allow_pickle=True).tolist()
    catid2color = dict()
    img = cv2.imread(image)
    img, scale_value = resize_img(img)
    color_map = get_color_map_list(len(labels) + 1)
    img_h = anno['h']
    img_w = anno['w']
    gt_class = anno['gt_class']
    gt_bbox = anno['gt_bbox']
    gt_poly = anno['gt_poly']
    num_bbox = gt_bbox.shape[0]
    num_mask = len(gt_poly)
    # 描绘mask信息
    img_array = np.array(img).astype('float32')
    for i in range(num_mask):
        cname = gt_class[i]
        catid = labels.index(cname)
        if cname not in labels:
            continue
        if catid not in catid2color:
            catid2color[catid] = color_map[catid + 1]
        color = np.array(catid2color[catid]).astype('float32')

        import pycocotools.mask as mask_util
        for x in range(len(gt_poly[i])):
            for y in range(len(gt_poly[i][x])):
                gt_poly[i][x][y] = int(float(gt_poly[i][x][y]) * scale_value)
        poly = gt_poly[i]
        rles = mask_util.frPyObjects(poly,
                                     int(float(img_h) * scale_value),
                                     int(float(img_w) * scale_value))
        rle = mask_util.merge(rles)
        mask = mask_util.decode(rle) * 255
        idx = np.nonzero(mask)
        img_array[idx[0], idx[1], :] *= 1.0 - alpha
        img_array[idx[0], idx[1], :] += alpha * color
    img = img_array.astype('uint8')

    for i in range(num_bbox):
        cname = gt_class[i]
        catid = labels.index(cname)
        if cname not in labels:
            continue
        if catid not in catid2color:
            catid2color[catid] = color_map[catid]
        color = tuple(catid2color[catid])
        xmin, ymin, xmax, ymax = gt_bbox[i]

        img = draw_rectangle_and_cname(img,
                                       int(float(xmin) * scale_value),
                                       int(float(ymin) * scale_value),
                                       int(float(xmax) * scale_value),
                                       int(float(ymax) * scale_value), cname,
                                       color)

    return img


def draw_rectangle_and_cname(img, xmin, ymin, xmax, ymax, cname, color):
    """ 根据提供的标注信息，给图片描绘框体和类别显示

    Args:
        img: 图片路径
        xmin: 检测框最小的x坐标
        ymin: 检测框最小的y坐标
        xmax: 检测框最大的x坐标
        ymax: 检测框最大的y坐标
        cname: 类别信息
        color: 类别与颜色的对应信息
    """
    # 描绘检测框
    line_width = math.ceil(2 * max(img.shape[0:2]) / 600)
    cv2.rectangle(
        img,
        pt1=(xmin, ymin),
        pt2=(xmax, ymax),
        color=color,
        thickness=line_width)

    # 计算并描绘类别信息
    text_thickness = math.ceil(2 * max(img.shape[0:2]) / 1200)
    fontscale = math.ceil(0.5 * max(img.shape[0:2]) / 600)
    tw, th = cv2.getTextSize(
        cname, 0, fontScale=fontscale, thickness=text_thickness)[0]
    cv2.rectangle(
        img,
        pt1=(xmin + 1, ymin - th),
        pt2=(xmin + int(0.7 * tw) + 1, ymin),
        color=color,
        thickness=-1)
    cv2.putText(
        img,
        cname, (int(xmin) + 3, int(ymin) - 5),
        0,
        0.6 * fontscale, (255, 255, 255),
        lineType=cv2.LINE_AA,
        thickness=text_thickness)
    return img


def pil_imread(file_path):
    """ 将图片读成np格式数据

    Args:
        file_path: 图片路径
    """
    img = Image.open(file_path)
    return np.asarray(img)


def get_color_map_list(num_classes):
    """ 为类别信息生成对应的颜色列表

    Args:
        num_classes: 类别数量
    """
    color_map = num_classes * [0, 0, 0]
    for i in range(0, num_classes):
        j = 0
        lab = i
        while lab:
            color_map[i * 3] |= (((lab >> 0) & 1) << (7 - j))
            color_map[i * 3 + 1] |= (((lab >> 1) & 1) << (7 - j))
            color_map[i * 3 + 2] |= (((lab >> 2) & 1) << (7 - j))
            j += 1
            lab >>= 3
    color_map = [color_map[i:i + 3] for i in range(0, len(color_map), 3)]
    return color_map


def gray2pseudo(gray_image):
    """ 将分割的结果映射到图片

    Args:
        gray_image: 灰度图
    """
    color_map = get_color_map_list(256)
    color_map = np.array(color_map).astype("uint8")
    # 用OpenCV进行色彩映射
    c1 = cv2.LUT(gray_image, color_map[:, 0])
    c2 = cv2.LUT(gray_image, color_map[:, 1])
    c3 = cv2.LUT(gray_image, color_map[:, 2])
    pseudo_img = np.dstack((c1, c2, c3))
    return pseudo_img
